import time


# Constants for the circuit breaker states
CLOSED = 1
OPEN = 2
HALF_OPEN = 3


class CircuitBreaker(object):

    def __init__(self, name, threshold=None, delay=None):
        """A circuit breaker opening when a failure threshold is reached.

        It half opens again when an attempt is made after a delay.
        It then closes again when a success is seen.

        :param threshold: When the number of failures exceeds it, the breaker
            is opened.

        :param delay: Number of seconds before attempting to close (so-called
            half open) the breaker.
        """
        if threshold is None:
            # No threshold
            threshold = 0
        if delay is None:
            delay = 0.0
        self.name = name
        self.failures = 0
        self.threshold = threshold
        self.delay = delay # in secs
        self.state = None
        # Assume it works
        self._close()

    def closed(self):
        """The circuit works if the breaker is closed."""
        return self.state == CLOSED

    def see_failure(self):
        # One more failure
        self.failures += 1
        if self.state == CLOSED:
            self._open()
        elif self.state == OPEN:
            self._half_open()
        elif self.state == HALF_OPEN:
            # Fail again
            self._open()
        else:
            raise AssertionError('Unkown circuit breaker state: %s'
                                 % (self.state,))

    def see_success(self):
        if self.state == CLOSED:
            # Everything works
            pass
        elif self.state == OPEN:
            # Weird, this is not expected in the normal workflow.  We assume a
            # wrong ordering in the events (or too many of them) but move to
            # half_open first, waiting for a confirmation.
            self._half_open()
        elif self.state == HALF_OPEN:
            self._close()
        else:
            raise AssertionError('Unkown circuit breaker state: %s'
                                 % (self.state,))

    def see_attempt(self):
        """An attempt to use the circuit is made.

        :returns: False if the circuit is open, True otherwise. This represents
            whether or not the attempt tried to use the circuit.
        """
        if self.state == CLOSED:
            # Everything works.
            pass
        elif self.state == OPEN:
            self._half_open()
        elif self.state == HALF_OPEN:
            # We're trying. Trying more than once in parallel is allowed.
            pass
        else:
            raise AssertionError('Unkown circuit breaker state: %s'
                                 % (self.state,))
        # The attempt is succesful if the breaker is not opened
        return not (self.state == OPEN)

    def _close(self):
        """The circuit works, close the breaker."""
        self.state = CLOSED
        self.failures = 0

    def _open(self):
        """A failure happened, open the breaker when reaching the threshold."""
        if self.failures > self.threshold:
            # Enough failures, we have a problem
            self.state = OPEN
            self.opened_at = time.time()

    def _half_open(self):
        """We're trying to close the breaker after a failure."""
        now = time.time()
        when = self.opened_at + self.delay
        if  now >= when:
            # We waited long enough, let's try again.
            self.state = HALF_OPEN

