from datetime import datetime, timedelta
import errno
import fcntl
from itertools import chain
import operator
import os
import re
import shutil
import StringIO
import time
import urllib2
import urlparse

try:
    from debian import changelog
except ImportError:
    from debian_bundle import changelog

from bzrlib import (
    branch,
    bzrdir,
    cleanup,
    delta,
    diff,
    errors,
    merge,
    tag,
    testament,
    transport,
    )
from bzrlib.trace import mutter
from launchpadlib.errors import HTTPError

from udd import (
    iconfig,
    idb,
    lpapi,
    )


# Not ideal to declare this here but this is used by several local functions
# below. Some way to share the config is needed -- vila 2011-12-07
conf = iconfig.ImporterStack()

sqlite_timeout = 30
running_sentinel = "Apparently the supervisor died\n"
no_lock_returncode = 139


def ensure_directory(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


urllib_launchpad_base_url = "https+urllib://launchpad.net/"
urllib_debian_base_url = "http+urllib://ftp.uk.debian.org/debian/"
urllib_debian_archive_base_url = "http+urllib://archive.debian.org/debian/"

# Distro releases, in chronological order, with continuing trunks like 'sid'
# last, as used to set the order of branches in a DistributionBranchSet.
db_set_distro_releases = {
    u"ubuntu": [u"warty", u"hoary", u"breezy",
               u"dapper", u"edgy", u"feisty", u"gutsy",
               u"hardy", u"intrepid", u"jaunty", u"karmic",
               u"lucid", u"maverick", u"natty", u"oneiric", u"precise",],
    u"debian": [u"woody", u"sarge", u"etch", u"lenny",
               u"squeeze", u"wheezy",
               u"sid", u"experimental"],
}

# Distro releases, in an order such that whenever the _same_ version of a
# package is published in multiple releases, it is published first in releases
# earlier in the list. Examples include, in Ubuntu, later releases being
# initialized as a copy of their parent (so later Ubuntu releases come after
# earlier ones here), and in Debian, packages propagating from sid to testing
# (hence sid comes before all suites that have ever been testing). This
# sequencing ensures that the importer produces the same structure of imports
# and pulls from other branches when run from scratch as when run incrementally
# after each publishing occurred.
import_sequence_distro_releases = {
    u"ubuntu": [u"warty", u"hoary", u"breezy",
               u"dapper", u"edgy", u"feisty", u"gutsy",
               u"hardy", u"intrepid", u"jaunty", u"karmic",
               u"lucid", u"maverick", u"natty", u"oneiric", u"precise",],
    u"debian": [u"sid", u"experimental",
               u"woody", u"sarge", u"etch", u"lenny",
               u"squeeze", u"wheezy"],
}

# Distro releases known to Launchpad, in the order they should be processed: do
# the likely development focus first to fit better with stacking.  No etch or
# earlier as they aren't on lp
lp_distro_releases = {
    u"ubuntu": [u"precise", u"warty", u"hoary", u"breezy",
               u"dapper", u"edgy", u"feisty", u"gutsy",
               u"hardy", u"intrepid", u"jaunty", u"karmic",
               u"lucid", u"maverick", u"natty", u"oneiric",],
    u"debian": [u"sid", u"experimental",
               u"wheezy", u"squeeze", u"lenny"],
}

distro_pockets = {
    u"ubuntu": [u"release", u"security", u"proposed", u"updates", u"backports"],
    u"debian": [u"release",],
}

default_debian_diff_release = u"sid"
default_ubuntu_merge_source = u"squeeze"


def debian_base_url(release):
    if release in [u"hamm", u"slink", u"potato", u"woody", u"sarge", u"etch"]:
        return urllib_debian_archive_base_url
    return urllib_debian_base_url


def make_suite(release, pocket):
    if pocket == u"release":
        suite = release
    elif pocket in (u"security", u"proposed", u"updates", u"backports"):
        suite = u"%s-%s" % (release, pocket)
    else:
        assert False, "Unknown pocket: %s" % pocket
    return suite


def lock_path(dir_name, file_name):
    ensure_directory(dir_name)
    path = os.path.join(dir_name, file_name)
    f = open(path, "wb")
    try:
        fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return f
    except IOError, e:
        if e.errno in (errno.EACCES, errno.EAGAIN):
            return None
        raise


class PackageToImport(object):
    """This object is analogous to (in Soyuz terminology) a source package
    publication."""

    def __init__(self, name, version, distro, release, pocket, component="",
            on_lp=True, url=None):
        self.name = name
        self.version = version
        self.distro = distro
        self.release = release
        self.component = component
        self.pocket = pocket
        self.on_lp = on_lp
        self.url = url

    def __repr__(self):
        return ('%s(%s %s %s/%s/%s, c=%s, url=%s)'
                % (self.__class__.__name__, self.name, self.version,
                   self.distro, self.release, self.pocket, self.component,
                   self.url))

    def __str__(self):
        return "PackageToImport(%s, %s, %s, %s, %s, %s)" % (self.name,
                self.version, self.distro, self.release, self.pocket,
                self.component)

    def same_as(self, other):
        return (self.name == other.name
                and self.version == other.version
                and self.distro == other.distro
                and self.release == other.release
                and self.pocket == other.pocket)

    def get_url(self):
        if self.url is not None:
            return self.url
        if self.version.debian_version is not None:
            version_str = "%s-%s" % (self.version.upstream_version,
                    self.version.debian_version)
        else:
            version_str = self.version.upstream_version
        if self.distro == u"ubuntu":
            files_url = (urllib_launchpad_base_url
                         + "ubuntu/%s/+source/%s/%s/+files/"
                         % (self.release, self.name, str(self.version)))
        else:
            if self.on_lp:
                files_url = (urllib_launchpad_base_url
                             + "debian/%s/+source/%s/%s/+files/"
                             % (self.release, self.name, str(self.version)))
            else:
                start = self.name[0]
                if self.name.startswith("lib"):
                    start = self.name[:4]
                files_url = (debian_base_url(self.release)
                             + "pool/%s/%s/%s/"
                             % (self.component, start, self.name))
        self.url = files_url + "%s_%s.dsc" % (self.name, version_str)
        return self.url

    @staticmethod
    def import_order_comparator(a, b):
        """This comparator aims to sort prospective import jobs such that
        package versions are imported after versions they were derived from,
        and publications that are likely to be copied into other archives,
        series or pockets are imported before imports to the likely
        destinations, such that the source imported version is available to be
        pulled."""
        # Core assumption: packages always have higher versions than their
        # ancestors... except for backports.  Therefore, sort backports after
        # all non-backports.
        if a.pocket == u"backports" and b.pocket != u"backports":
            return 1
        if a.pocket != u"backports" and b.pocket == u"backports":
            return -1
        # Next sort by version.
        if a.version < b.version:
            return -1
        if a.version > b.version:
            return 1
        # From here on, a.version == b.version
        # Sort Debian before Ubuntu, so that synced packages are first imported
        # into Debian.
        if a.distro != b.distro:
            if a.distro == u"debian":
                return -1
            return 1
        # From here on, a.distro == b.distro
        # Release series should be imported in ascending order, modelling a
        # packaging being carried forward unchanged into new series.
        if a.release != b.release:
            a_release = import_sequence_distro_releases[a.distro].index(a.release)
            b_release = import_sequence_distro_releases[b.distro].index(b.release)
            if a_release < b_release:
                return -1
            return 1
        # Within a single release series, import pockets such that for all
        # typical pocket copy workflows, the source pocket is imported first.
        if a.pocket != b.pocket:
            a_pocket = distro_pockets[a.distro].index(a.pocket)
            b_pocket = distro_pockets[b.distro].index(b.pocket)
            if a_pocket < b_pocket:
                return -1
            return 1
        # Same version, distro, release and pocket
        return 0


class ImportList(object):

    def __init__(self):
        self.plist = []

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.plist)

    def sort(self):
        self.plist.sort(cmp=PackageToImport.import_order_comparator)

    def add_if_needed(self, newp):
        for oldp in self.plist:
            if oldp.same_as(newp):
                return
        self.plist.append(newp)

    def __str__(self):
        return "[" + ", ".join(map(str, self.plist)) + "]"


def get_last_import_time(status_db, earliest_date=None):
    """Get the last import time from status db.

    Adjusts for safety, so that we don't miss any packages on the new
    synchronization. If ``earliest_date`` provided, then never return a date
    earlier than that, even if we've never synced before.
    """
    last_known_published = status_db.last_import_time()
    if last_known_published:
        # Safety buffer so we don't accidentally miss other syncs.
        return last_known_published - timedelta(minutes=10)
    return earliest_date


def create_import_jobs(lp, status_db):
    # When starting from scratch, limit the query to the last ten days or
    # lp timeout (https://bugs.launchpad.net/udd/+bug/717109)
    last_known_published = get_last_import_time(
        status_db, earliest_date=datetime.now() - timedelta(days=10))
    # We want all the new Published records since the last run
    distributions = conf.get('pi.distributions')
    filters = {'status': 'Published',
               'since': last_known_published.isoformat(),
               }
    publications = []
    for distro in distributions:
        publications.extend(
            iter_publications(lp.distributions[distro], **filters))
    newest_published, checked = prepare_for_database(
        last_known_published, set(publications))
    status_db.add_import_jobs(checked, newest_published)


def prepare_for_database(last_known_published, publications):
    """Prepare data from Launchpad for the import job database.

    Given a last known published date and an iterable of ``Publication``s,
    return a new last known published date and a set of names.
    """
    newest_published = last_known_published
    checked = set()
    for publication in publications:
        checked.add(publication.name)
        published_time = publication.published_time
        if (newest_published is None or published_time > newest_published):
            newest_published = published_time
    return newest_published, checked


def iter_publications(distribution, status='Published', since=None,
                      series=None):
    """Iterate over published packages.

    Determines whether we want source or binary packages from the
    configuration.

    :param distribution: A Launchpad ``Distribution`` object. e.g. ubuntu.
    :param status: One of 'Pending', 'Published', 'Superseded', 'Deleted'
        or 'Obsolete'.  Defaults to 'Published'.
    :param since: If provided, only find packages published since that date.
    :param series: The name or version of the distribution series to restrict
        the search to. e.g. 'oneiric'
    :return: An iterator over `Publication` objects.
    """
    if conf.get('pi.scan_mode') == 'binary':
        iter_pubs = iter_published_binaries
    else:
        iter_pubs = iter_published_sources
    collection = iter_pubs(
        distribution, status=status, since=since, series=series)
    for pub in collection:
        yield lpapi.Publication.from_source_or_binary(pub)


def iter_published_sources(distribution, status='Published', since=None,
                           series=None):
    """Iterate over published source packages.

    :param distribution: A Launchpad ``Distribution`` object. e.g. ubuntu.
    :param status: One of 'Pending', 'Published', 'Superseded', 'Deleted'
        or 'Obsolete'.  Defaults to 'Published'.
    :param since: If provided, only find packages published since that date.
    :param series: The name or version of the distribution series to restrict
        the search to. e.g. 'oneiric'
    :return: An iterator over source package publishing history objects.
        See https://launchpad.net/+apidoc/devel.html#source_package_publishing_history
        for more details.
    """
    filters = {'status': status}
    if since:
        filters['created_since_date'] = since
    if series:
        series = distribution.getSeries(name_or_version=series)
        filters['distro_series'] = series
    archive = distribution.main_archive
    collection = lpapi.lp_call(
        lpapi.call_with_limited_size, archive.getPublishedSources, **filters)
    return lpapi.iterate_collection(collection)


def _lp_get_published_binaries(archive, distro_arch_series, **filters):
    """Actually make the call to Launchpad."""
    # This is a separate function so we can monkey-patch it and thus test the
    # code that calls it. The function itself is untested, and nigh on
    # untestable.
    return lpapi.iterate_collection(
        lpapi.lp_call(
            lpapi.call_with_limited_size, archive.getPublishedBinaries,
            distro_arch_series=distro_arch_series, **filters))


def iter_published_binaries(distribution, status='Published', since=None,
                            series=None):
    """Iterate over published binary packages.

    :param distribution: A Launchpad ``Distribution`` object. e.g. ubuntu.
    :param status: One of 'Pending', 'Published', 'Superseded', 'Deleted'
        or 'Obsolete'.  Defaults to 'Published'.
    :param since: If provided, only find packages published since that date.
    :param series: If provided, only find packages published to that series.
    :return: An iterator over source package publishing history objects.
        See https://launchpad.net/+apidoc/devel.html#binary_package_publishing_history
        for more details.
    """
    filters = {'status': status}
    if since:
        filters['created_since_date'] = since
    archive = distribution.main_archive
    if series is None:
        # If series is None, then a single, unadorned getPublishedBinaries
        # query will time out (bug #904344).  Instead, break down the query
        # into one for each series and then combine the results.
        collections = []
        for s in distribution.series:
            collections.append(
                iter_published_binaries(distribution, status, since, s))
        return chain(*collections)
    collections = []
    # There's no way to query getPublishedBinaries by distro_series, so we
    # have to search by distro_arch_series and combine the results.
    # <https://bugs.launchpad.net/launchpad/+bug/904670>.
    for distro_arch_series in series.architectures:
        collections.append(
            _lp_get_published_binaries(archive, distro_arch_series, **filters))
    return chain(*collections)


class PackageStatus(object):

    def __init__(self, name):
        self.name = name
        self.failure_count = 0
        self.running = False
        self.queued = False
        self.auto_retry = False
        self.auto_retry_masked = False
        self.timestamp = None
        self.auto_retry_time = None
        self.signature = None
        self.raw_reason = None


class Cursor(object):
    """A cursor as a context manager that commits automatically."""

    def __init__(self, connection):
        # Capture the connection we'll be using
        self.conn = connection

    def __enter__(self):
        # Create a cursor
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            # All went well, commit the changes
            self.conn.commit()
            return True
        # Something went wrong, rollback the changes
        self.conn.rollback()
        return False


class StatusDatabase(object):
    """General metadata database, typically in meta.db"""

    FAILURES_TABLE = "failures"
    FAILURES_TABLE_CREATE = '''create table if not exists %s
                     (package text constraint nonull NOT NULL,
                      reason blob,
                      when_failed timestamp,
                      emailed integer default 0,
                      constraint isprimary PRIMARY KEY
                            (package))''' % FAILURES_TABLE
    FAILURES_TABLE_FIND = 'select * from %s where package=?' % FAILURES_TABLE
    FAILURES_TABLE_DELETE = 'delete from %s where package=?' % FAILURES_TABLE

    OLD_FAILURES_TABLE = "old_failures"
    OLD_FAILURES_TABLE_CREATE = '''create table if not exists %s
                     (package text constraint nonull NOT NULL,
                      reason blob,
                      when_failed timestamp,
                      last_failed timestamp,
                      failure_count integer,
                      constraint isprimary PRIMARY KEY
                            (package))''' % OLD_FAILURES_TABLE
    OLD_FAILURES_TABLE_FIND = '''select * from %s where package=?''' % OLD_FAILURES_TABLE
    OLD_FAILURES_TABLE_DELETE = '''delete from %s where package=?''' % OLD_FAILURES_TABLE

    JOBS_TABLE = "jobs"
    JOBS_TABLE_CREATE = '''create table if not exists %s
                    (id integer primary key asc,
                     package text constraint nonull NOT NULL,
                     active integer constraint nonull NOT NULL,
                     type integer constraint nonull NOT NULL,
                     date_requested timestamp constraint nonull NOT NULL,
                     date_started timestamp,
                     date_completed timestamp)''' % JOBS_TABLE
    JOBS_TABLE_FIND = '''select * from %s where package=?
                    and active=? order by type desc''' % JOBS_TABLE
    JOBS_TABLE_NEXT = '''select * from %s where active=1
                    order by type desc''' % JOBS_TABLE
    JOBS_TABLE_SELECT_JOB = '''select * from %s where id=?''' % JOBS_TABLE
    JOBS_TABLE_INSERT = '''insert into %s values (?, ?, ?, ?, ?, ?, ?)''' % JOBS_TABLE
    JOBS_TABLE_UPDATE = '''update %s set package=?, active=?, type=?,
                    date_requested=?, date_started=?, date_completed=?
                    where id=?''' % JOBS_TABLE

    IMPORT_TABLE = "import"
    IMPORT_TABLE_CREATE = '''create table if not exists %s
                    (import timestamp constraint nonull NOT NULL)
                    ''' % IMPORT_TABLE

    RETRY_TABLE = "should_retry"
    RETRY_TABLE_CREATE = '''create table if not exists %s
                    (signature blob constraint nonull NOT NULL,
                     constraint isprimary PRIMARY KEY
                            (signature))''' % RETRY_TABLE
    RETRY_TABLE_SELECT = '''select * from %s where signature=?''' % RETRY_TABLE
    RETRY_TABLE_INSERT = '''insert into %s values (?)''' % RETRY_TABLE

    JOB_TYPE_PRIORITY = 5
    JOB_TYPE_NEW = 3
    JOB_TYPE_RETRY = 1

    AUTO_RETRY_SECONDS = 60*60*3
    MAX_AUTO_RETRY_COUNT = 5

    def __init__(self, db_conn, db_type):
        self.conn = db_conn
        with Cursor(self.conn) as c:
            c.execute(self.FAILURES_TABLE_CREATE)
            c.execute(self.OLD_FAILURES_TABLE_CREATE)
            c.execute(self.JOBS_TABLE_CREATE)
            c.execute(self.IMPORT_TABLE_CREATE)
            c.execute(self.RETRY_TABLE_CREATE)

    def _has_failed(self, c, package):
        rows = list(c.execute(self.FAILURES_TABLE_FIND, (package,)))
        return (len(rows) > 0)

    def _set_failure(self, c, package, reason, now):
        if self._has_failed(c, package):
            c.execute('update %s set package=?, reason=?, when_failed=? '
                    'where package=?' % self.FAILURES_TABLE,
                    (package, reason, now, package))
        else:
            c.execute('insert into %s values (?, ?, ?, 0)'
                    % self.FAILURES_TABLE, (package, reason, now))


    def add_failed_jobs(self):
        """Add all failed jobs to the queue."""
        with Cursor(self.conn) as c:
            rows = c.execute('select * from %s' % self.FAILURES_TABLE
                             )
            for row in rows:
                c.execute(self.FAILURES_TABLE_DELETE, (row[0],))
                self._add_job(c, row[0], self.JOB_TYPE_PRIORITY)

    def add_jobs_for_interrupted(self):
        with Cursor(self.conn) as c:
            rows = c.execute('select * from %s where reason=?'
                    % self.FAILURES_TABLE, (running_sentinel,))
            for row in rows:
                c.execute(self.FAILURES_TABLE_DELETE, (row[0],))
                self._add_job(c, row[0], self.JOB_TYPE_PRIORITY)

    def _start_package(self, c, package):
        ret = None
        # Search the active jobs for the given package ordered by job type
        # (priority, new, retry)
        rows = c.execute(self.JOBS_TABLE_FIND, (package, 1))
        now = datetime.utcnow()
        first = now
        job_id = 0
        for row in rows:
            # Mark all the jobs inactive (and only the first as "started")
            c.execute(self.JOBS_TABLE_UPDATE,
                    (row[1], 0, row[3], row[4], first, row[6], row[0]))
            first = None
            job_id = row[0]
        if self._has_failed(c, package):
            # Don't start it if it has failed
            job_id = None
        else:
            # Mark the job as one that will be started RSN
            c.execute('insert into %s values (?, ?, ?, 0)'
                      % self.FAILURES_TABLE,
                      (package, running_sentinel, now))
            ret = job_id
        return ret

    def start_package(self, package):
        with Cursor(self.conn) as c:
            return self._start_package(c, package)

    def finish_job(self, package, job_id, success, output, when=None):
        if when is None:
            when = datetime.utcnow()
        with Cursor(self.conn) as c:
            if job_id > 0:
                row = c.execute(self.JOBS_TABLE_SELECT_JOB,
                                (job_id,)).get_one()
                if row is not None:
                    c.execute(self.JOBS_TABLE_UPDATE,
                            (row[1], 0, row[3], row[4], row[5], when, row[0]))
            if success:
                row = c.execute(self.FAILURES_TABLE_FIND, (package,)).get_one()
                if row is not None:
                    c.execute('delete from %s where package=?'
                            % self.JOBS_TABLE, (package,))
                c.execute(self.FAILURES_TABLE_DELETE, (package,))
                c.execute(self.OLD_FAILURES_TABLE_DELETE, (package,))
            else:
                self._set_failure(c, package, output, when)

    def _add_job(self, c, package, job_type):
        c.execute(self.JOBS_TABLE_INSERT, (None, package, 1, job_type,
                    datetime.utcnow(), None, None))

    def add_jobs(self, packages, job_type):
        with Cursor(self.conn) as c:
            for package in packages:
                self._add_job(c, package, job_type)

    def last_import_time(self):
        with Cursor(self.conn) as c:
            row = c.execute('select * from %s order by import desc'
                    % self.IMPORT_TABLE).get_one()
            if row is not None:
                return make_datetime(row[0])
            return row

    def add_import_jobs(self, packages, newest_published):
        with Cursor(self.conn) as c:
            for package in packages:
                self._add_job(c, package, self.JOB_TYPE_NEW)
            c.execute('insert into %s values (?)' % self.IMPORT_TABLE,
                    (newest_published,))

    def next_job(self):
        with Cursor(self.conn) as c:
            job_id = 0
            package = None
            found = False
            while not found:
                row = c.execute(self.JOBS_TABLE_NEXT).get_one()
                if row is not None:
                    package = row[1]
                    job_id = self._start_package(c, package)
                    if job_id is None:
                        c.execute(self.JOBS_TABLE_UPDATE,
                                (row[1], 0, row[3], row[4], row[5],
                                 row[6], row[0]))
                    else:
                        found = True
                else:
                    found = True
                    job_id = 0
                    package = None
            return (job_id, package)

    def failure_reason(self, package):
        with Cursor(self.conn) as c:
            row = c.execute(self.FAILURES_TABLE_FIND, (package,)).get_one()
            if row is None:
                return None
            return row[1]

    def failure_signature(self, raw_reason):
        trace = raw_reason.splitlines()
        if len(trace) == 1:
            if trace[0] == running_sentinel[:-1]: # Get rid of the final '\n'
                return None
            # sometimes, Python exceptions do not have file references
            m = re.match('(\w+): ', trace[0])
            if m:
                return m.group(1)
            else:
                return trace[0].strip()
        elif len(trace) < 3:
            return " ".join(trace).strip()

        # If the failure reason is a traceback (which should always be the
        # case, the running_sentinel check above taking care of the still
        # running imports), we build the traceback signature and capture the
        # exception type
        sig = ''
        exc_type = ''
        exc_type_coming = False
        file_line_seen = False
        for l in trace:
            if l.startswith('  File'):
                # Keep the method/function name
                sig += ':' + l.split()[-1]
                file_line_seen = True
            elif file_line_seen:
                # We've seen the 'File...' line, so we are now seeing the code
                # line
                exc_type_coming = True
                file_line_seen = False
            elif exc_type_coming:
                # We've seen the code line so we may find the exception line
                # itself now.

                # We have no way to know if we are at the last line of the
                # traceback, but we can't rely on the execption always
                # displaying a single line either, so the last captured
                # exception will do.
                exc_type = l.split(':')[0]
                exc_type_coming = False

        sig = exc_type + sig
        return sig

    def _known_auto_retry(self, c, sig):
        return c.execute(self.RETRY_TABLE_SELECT, (sig,)).get_one()

    def known_auto_retry(self, sig):
        with Cursor(self.conn) as c:
            return self._known_auto_retry(c, sig)

    def retry(self, package, force=False, priority=False, auto=False,
            all=False):
        with Cursor(self.conn) as c:
            row = c.execute(self.FAILURES_TABLE_FIND, (package,)).get_one()
            if row is None:
                if not force:
                    return False
                job_type = self.JOB_TYPE_RETRY
                if priority:
                    job_type = self.JOB_TYPE_PRIORITY
                self._add_job(c, package, job_type)
            else:
                raw_reason = row[1]
                raw_reason = raw_reason.encode("ascii", "replace")
                sig = self.failure_signature(raw_reason)
                self._retry(c, package, sig, row[2], priority=priority)
                if auto and sig != None:
                    row = self._known_auto_retry(c, sig)
                    if row is None:
                        c.execute(self.RETRY_TABLE_INSERT, (sig,))
                if all and sig != None:
                    rows = c.execute('select * from %s'
                            % self.FAILURES_TABLE)
                    for row in rows:
                        this_raw_reason = row[1].encode("ascii", "replace")
                        this_sig = self.failure_signature(this_raw_reason)
                        if this_sig == sig:
                            self._retry(c, row[0], sig, row[2],
                                    priority=priority)
            return True

    def _retry(self, c, package, signature, timestamp, priority=False):
        row = c.execute(self.OLD_FAILURES_TABLE_FIND, (package,)).get_one()
        if row is not None:
            if signature == row[1]:
                if row[4] > self.MAX_AUTO_RETRY_COUNT:
                    print ("Warning: %s has failed %d times in the same way"
                            % (package, row[4]))
                failure_count = row[4]+1
            else:
                failure_count = 1
            c.execute('update %s set package=?, reason=?, when_failed=?, '
                      'last_failed=?, failure_count=? where package=?'
                      % self.OLD_FAILURES_TABLE,
                      (package, signature, timestamp, timestamp, failure_count,
                       package))
        else:
            c.execute('insert into %s values (?, ?, ?, ?, ?)'
                    % self.OLD_FAILURES_TABLE,
                    (package, signature, timestamp, timestamp, 1))
        c.execute(self.FAILURES_TABLE_DELETE, (package,))
        job_type = self.JOB_TYPE_RETRY
        if priority:
            job_type = self.JOB_TYPE_PRIORITY
        self._add_job(c, package, job_type)

    def _attempt_retry(self, c, info):
        row = self._known_auto_retry(c, info.signature)
        if row is None:
            # A failure that is not auto-retried
            return False
        info.auto_retry = True
        info.auto_retry_time = (
            info.timestamp + timedelta(seconds=self.AUTO_RETRY_SECONDS))
        if info.auto_retry_time > datetime.utcnow():
            # Too early to retry
            return False
        row = c.execute(self.OLD_FAILURES_TABLE_FIND, (info.name,)).get_one()
        if row is not None and row[1] == info.signature:
            info.failure_count += row[4]
            if row[4] > self.MAX_AUTO_RETRY_COUNT:
                info.auto_retry_masked = True
                # Already retried MAX_AUTO_RETRY_COUNT times, stop retrying
                return False
        self._retry(c, info.name, info.signature, info.timestamp)

    def summarise_failures(self):
        package_info = []
        with Cursor(self.conn) as c:
            rows = c.execute('select * from %s'
                    % self.FAILURES_TABLE)
            db_failures = [(row[0], row[1], row[2]) for row in rows]
            seen_package_names = set()
            for row in c.execute(self.JOBS_TABLE_NEXT):
                if row[1] in seen_package_names:
                    continue
                seen_package_names.add(row[1])
                package_info.append(PackageStatus(row[1]))
                package_info[-1].queued = True
            del seen_package_names
            reasons = {}
            for package, raw_reason, ts in db_failures:
                package_info.append(PackageStatus(package))
                if raw_reason == running_sentinel:
                    package_info[-1].running = True
                    package_info[-1].timestamp = ts
                    continue
                raw_reason = raw_reason.encode("ascii", "replace")
                sig = self.failure_signature(raw_reason)
                if sig is None:
                    continue
                else:
                    info = package_info[-1]
                    info.failure_count = 1
                    info.signature = sig
                    info.raw_reason = raw_reason
                    info.timestamp = ts
                    if self._attempt_retry(c, info):
                        continue
                    reasons.setdefault(sig, [])
                    reasons[sig].append(info)
            return (reasons, package_info)

    def unemailed_failures(self):
        with Cursor(self.conn) as c:
            rows = c.execute('select * from %s where emailed=0'
                    % self.FAILURES_TABLE)
            db_failures = [(row[0], row[1], make_datetime(row[2])) for row in rows]
            return db_failures

    def set_failures_emailed(self, failures):
        with Cursor(self.conn) as c:
            for failure in failures:
                c.execute('update %s set emailed=1 where package=? '
                        '' % (self.FAILURES_TABLE),
                        (failure[0], ))


def make_datetime(strified):
    fmt = "%Y-%m-%d %H:%M:%S"
    if "." in strified:
        fmt += ".%f"
    return datetime.strptime(strified, fmt)


class CommitDatabase(object):

    COMMIT_TABLE = "commits"
    COMMIT_TABLE_CREATE = '''create table if not exists %s
                     (package text constraint nonull NOT NULL,
                      constraint isprimary PRIMARY KEY
                            (package))''' % COMMIT_TABLE
    COMMIT_TABLE_FIND = '''select * from %s where package=?''' % COMMIT_TABLE
    COMMIT_TABLE_INSERT = '''insert into %s values (?)''' % COMMIT_TABLE
    COMMIT_TABLE_DELETE = "delete from %s where package=?" % COMMIT_TABLE

    def __init__(self, db_conn, db_type, package):
        self.package = package
        self.conn = db_conn
        with Cursor(self.conn) as c:
            c.execute(self.COMMIT_TABLE_CREATE)

    def has_commit_started(self):
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.COMMIT_TABLE_FIND, (self.package,)))
            return len(rows) > 0

    def start_commit(self):
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.COMMIT_TABLE_FIND, (self.package,)))
            assert len(rows) == 0, "Already started commit"
            c.execute(self.COMMIT_TABLE_INSERT, (self.package,))

    def finish_commit(self):
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.COMMIT_TABLE_FIND, (self.package,)))
            assert len(rows) == 1, "Commit not started"
            c.execute(self.COMMIT_TABLE_DELETE, (self.package,))


class PackageDatabase(object):
    """Database holding a list of known packages and whether they're in main or not.

    Typically in packages.db.
    """

    PACKAGE_TABLE = "packages"
    PACKAGE_TABLE_CREATE = '''create table if not exists %s
                     (package text constraint nonull NOT NULL,
                      main integer constraint nonull NOT NULL,
                      constraint isprimary PRIMARY KEY
                            (package))''' % PACKAGE_TABLE
    PACKAGE_TABLE_FIND = '''select * from %s where package=?''' % PACKAGE_TABLE
    PACKAGE_TABLE_INSERT = '''insert into %s values (?, ?)''' % PACKAGE_TABLE
    PACKAGE_TABLE_UPDATE = '''update %s set package=?, main=? where
                    package=?''' % PACKAGE_TABLE
    PACKAGE_TABLE_FIND_MAIN = '''select package from %s where main=1''' % PACKAGE_TABLE
    PACKAGE_TABLE_FIND_UNIVERSE = '''select package from %s where main=0''' % PACKAGE_TABLE

    UPDATE_TABLE = "packages_update"
    UPDATE_TABLE_CREATE = '''create table if not exists %s
                    (anupdate timestamp constraint nonull NOT NULL)
                    ''' % UPDATE_TABLE

    def __init__(self, db_conn, db_type):
        self.conn = db_conn
        with Cursor(self.conn) as c:
            c.execute(self.PACKAGE_TABLE_CREATE)
            c.execute(self.UPDATE_TABLE_CREATE)

    def last_update(self):
        with Cursor(self.conn) as c:
            row = c.execute('select * from %s order by anupdate desc'
                    % self.UPDATE_TABLE).get_one()
            if row is not None:
                return row[0]
            return row

    def update_packages(self, package_dict):
        with Cursor(self.conn) as c:
            for (package, main) in package_dict.items():
                if len(list(c.execute(self.PACKAGE_TABLE_FIND, (package,)
                            ))) > 0:
                    c.execute(self.PACKAGE_TABLE_UPDATE,
                            (package, main, package))
                else:
                    c.execute(self.PACKAGE_TABLE_INSERT, (package, main))
            c.execute('insert into %s values (?)' % self.UPDATE_TABLE,
                    (datetime.utcnow(),))

    def list_packages_in_main(self):
        with Cursor(self.conn) as c:
            rows = c.execute(self.PACKAGE_TABLE_FIND_MAIN)
            return set([row[0] for row in rows])

    def get_one(self, these_ones):
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.PACKAGE_TABLE_FIND_MAIN))
            count = len(rows)
            for row in rows:
                if row[0] not in these_ones:
                    return row[0]
            rows = list(c.execute(self.PACKAGE_TABLE_FIND_UNIVERSE))
            count += len(rows)
            if count == 0:
                raise StopIteration
            for row in rows:
                if row[0] not in these_ones:
                    return row[0]


class RevidDatabase(object):

    REVID_TABLE = "revids"
    REVID_TABLE_CREATE = '''create table if not exists %s
                     (package text constraint nonull NOT NULL,
                      version text constraint nonull NOT NULL,
                      suite text constraint nonull NOT NULL,
                      revid text constraint nonull NOT NULL,
                      testament text constraint nonull NOT NULL,
                      constraint isprimary PRIMARY KEY
                            (package, version, suite))'''
    REVID_TABLE_FIND = '''select * from %s where version=?
                          and package=? and suite=?'''
    REVID_TABLE_FIND_SUITE = '''select version from %s where package=?
                                and suite=?'''
    REVID_WORKING_TABLE = "revids_working"
    REVID_TABLE_FIND_PACKAGE = '''select * from %s where package=?'''
    REVID_TABLE_INSERT = "insert into %s values (?, ?, ?, ?, ?)"
    DELETE_WORKING = "delete from %s where package=?" % REVID_WORKING_TABLE

    SUFFIX_TABLE = "suffixes"
    SUFFIX_TABLE_CREATE = '''create table if not exists %s
                    (package text constraint nonull NOT NULL,
                     suffix text constraint nonull NOT NULL,
                     constraint isprimary PRIMARY KEY (package))''' % SUFFIX_TABLE


    def __init__(self, db_conn, db_type, package):
        self.package = package
        self.conn = db_conn
        with Cursor(self.conn) as c:
            c.execute(self.REVID_TABLE_CREATE % self.REVID_TABLE)
            c.execute(self.REVID_TABLE_CREATE % self.REVID_WORKING_TABLE)
            c.execute(self.SUFFIX_TABLE_CREATE)
        self.outstanding_marks = {}

    def _memory_rows(self):
        # Access to the internal state for tests
        return self.outstanding_marks.keys()

    def _working_rows(self):
        # Access to the internal state for tests
        with Cursor(self.conn) as c:
            return list(c.execute(self.REVID_TABLE_FIND_PACKAGE
                    % self.REVID_WORKING_TABLE, (self.package,)))

    def _saved_rows(self):
        # Access to the internal state for tests
        with Cursor(self.conn) as c:
            return list(c.execute(self.REVID_TABLE_FIND_PACKAGE
                    % self.REVID_TABLE, (self.package,)))

    def is_marked(self, version, suite):
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.REVID_TABLE_FIND % self.REVID_TABLE,
                    (unicode(version), self.package, suite)))
            rows += list(c.execute(self.REVID_TABLE_FIND % self.REVID_WORKING_TABLE,
                    (unicode(version), self.package, suite)))
            rows += [foo for foo in self.outstanding_marks
                        if foo == (unicode(version), suite)]
            assert len(rows) < 2, "Multiple versions for package/suite?"
            if len(rows) > 0:
                return True
            return False

    def last_marked_version(self, suite):
        with Cursor(self.conn) as c:
            versions = [a[0] for a in c.execute(self.REVID_TABLE_FIND_SUITE
                    % self.REVID_WORKING_TABLE,
                    (self.package, suite))]
            versions += [a[0] for a in c.execute(self.REVID_TABLE_FIND_SUITE
                    % self.REVID_TABLE,
                    (self.package, suite))]
            versions += [unicode(foo[0]) for foo in self.outstanding_marks
                            if foo[1] == suite]
            if len(versions) > 0:
                versions.sort(key=lambda x: changelog.Version(x))
                return versions[-1]
            return None

    def check(self, version, revid, suite, branch):
        sha = self.get_testament_sha1(branch.repository, revid)
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.REVID_TABLE_FIND % self.REVID_TABLE,
                    (unicode(version), self.package, suite)))
            rows += list(c.execute(self.REVID_TABLE_FIND % self.REVID_WORKING_TABLE,
                    (unicode(version), self.package, suite)))
            if (version, suite) in self.outstanding_marks:
                revid, tment = self.outstanding_marks[(unicode(version), suite)]
                rows += (None, None, None, revid, tment)
            if len(rows) < 1:
                return False
            assert len(rows) < 2, "Multiple versions for a package/suite?"
            row = rows[0]
            if (revid, sha) != (row[3], row[4]):
                assert False, ("%s != %s for %s %s in %s, something has changed"
                     % (unicode((revid, sha)), unicode((row['revid'],
                                 row['testament'])), self.package,
                         unicode(version), suite))
            return True

    def get_testament_sha1(self, repo, revid):
        op = cleanup.OperationWithCleanups(self._get_testament_sha1)
        return op.run(repo, revid)

    def _get_testament_sha1(self, op, repo, revid):
        op.add_cleanup(repo.lock_read().unlock)
        tment = testament.StrictTestament3.from_revision(repo, revid)
        return tment.as_sha1()

    def _commit(self, c):
        rows = list(c.execute(self.REVID_TABLE_FIND_PACKAGE
                % self.REVID_WORKING_TABLE, (self.package,)))
        if len(rows) > 0:
            def get_values():
                for row in rows:
                    yield tuple([row[i] for i in range(5)])
            for values in get_values():
                c.execute(self.REVID_TABLE_INSERT % self.REVID_TABLE,
                        values)
            c.execute(self.DELETE_WORKING, (self.package,))

    def commit(self):
        with Cursor(self.conn) as c:
            self._commit(c)

    def cleanup_last_run(self, cleanup_cb, push_cb):
        with Cursor(self.conn) as c:
            rows = list(c.execute(self.REVID_TABLE_FIND_PACKAGE
                    % self.REVID_WORKING_TABLE, (self.package,)))
            if len(rows) < 1:
                mutter("Nothing to cleanup")
            else:
                self.conn.rollback()
                push_cb()
                self._commit(c)
        cleanup_cb()

    def discard_last_run(self):
        with Cursor(self.conn) as c:
            c.execute(self.DELETE_WORKING, (self.package,))

    def mark(self, version, revid, suite, branch):
        sha = self.get_testament_sha1(branch.repository, revid)
        self.outstanding_marks[(unicode(version), suite)] = (revid, sha)

    def commit_outstanding(self):
        with Cursor(self.conn) as c:
            for version, suite in self.outstanding_marks:
                revid, sha = self.outstanding_marks[(version, suite)]
                rows = list(c.execute(self.REVID_TABLE_FIND % self.REVID_TABLE,
                        (unicode(version), self.package, suite)))
                rows += list(c.execute(self.REVID_TABLE_FIND % self.REVID_WORKING_TABLE,
                        (unicode(version), self.package, suite)))
                assert len(rows) < 1, "Trying to mark version %s in %s again" \
                    % (unicode(version), suite)
                c.execute(self.REVID_TABLE_INSERT % self.REVID_WORKING_TABLE,
                        (self.package, unicode(version), suite, revid, sha))
            self.outstanding_marks = {}

    def get_suffix(self):
        with Cursor(self.conn) as c:
            rows = list(c.execute("select * from %s where package=?"
                    % self.SUFFIX_TABLE, (self.package,)))
            if len(rows) < 1:
                return ""
            return rows[0][1]


class HistoryDatabase(object):
    """Records the data for a graph of queue and failures over time.

    Typically in history.db.
    """

    FAILED_TABLE = "failed"
    FAILED_TABLE_CREATE = '''create table if not exists %s
                     (attime timestamp constraint nonull NOT NULL,
                      queue integer constraint nonull NOT NULL,
                      failed integer constraint nonull NOT NULL,
                      constraint isprimary PRIMARY KEY (attime))'''
    MAIN_FAILED_TABLE = "main_failed"

    def __init__(self, db_conn, db_type):
        self.conn = db_conn
        with Cursor(self.conn) as c:
            c.execute(self.FAILED_TABLE_CREATE % self.FAILED_TABLE)
            c.execute(self.FAILED_TABLE_CREATE % self.MAIN_FAILED_TABLE)

    def _set_counts(self, table, queued, failed):
        with Cursor(self.conn) as c:
            c.execute("insert into %s values (?, ?, ?)" % table,
                      (datetime.utcnow(), queued, failed))

    def set_counts(self, queued, failed):
        self._set_counts(self.FAILED_TABLE, queued, failed)

    def set_main_counts(self, queued, failed):
        self._set_counts(self.MAIN_FAILED_TABLE, queued, failed)

    def _get_counts(self, table):
        with Cursor(self.conn) as c:
            rows = c.execute("select * from %s"
                    % table)
            return [(make_datetime(r[0]), r[1], r[2]) for r in rows]

    def get_main_counts(self):
        return self._get_counts(self.MAIN_FAILED_TABLE)

    def get_counts(self):
        return self._get_counts(self.FAILED_TABLE)


class MemoryRevidDatabase(RevidDatabase):

    def __init__(self, package):
        self.package = package
        self._maps = {}

    def is_marked(self, version, suite):
        revid_map = self._get_map(suite)
        return unicode(version) in revid_map

    def last_marked_version(self, suite):
        revid_map = self._get_map(suite)
        versions = revid_map.keys()
        if len(versions) > 0:
            versions.sort(key=lambda x: changelog.Version(x))
            return versions[-1]
        return None

    def check(self, version, revid, suite, branch):
        revid_map = self._get_map(suite)
        if not unicode(version) in revid_map:
            return False
        sha = self.get_testament_sha1(branch.repository, revid)
        if (revid, sha) != revid_map[unicode(version)]:
            mutter("%s != %s for %s %s, something has changed"
                   % (unicode((revid, sha)), unicode(revid_map[unicode(version)]),
                      self.package, unicode(version)))
            return False
        return True

    def commit(self):
        raise NotImplementedError(self.finish_commit)

    def cleanup_last_run(self, cleanup_cb, push_cb):
        pass

    def discard_last_run(self):
        pass

    def _get_map(self, suite):
        self._maps.setdefault(suite, {})
        return self._maps[suite]

    def _save_map(self, suite, map):
        self._maps[suite] = map

    def mark(self, version, revid, suite, branch):
        revid_map = self._get_map(suite)
        assert unicode(version) not in revid_map, \
              "Asked to mark %s %s %s, but it is already marked" \
                    % (self.package, version, suite)
        sha = self.get_testament_sha1(branch.repository, revid)
        revid_map[unicode(version)] = (revid, sha)
        self._save_map(suite, revid_map)

    def commit_outstanding(self):
        pass


# XXX: Move this and the other branch stuff into a separate module.
class BranchStore(object):

    def __init__(self, package, lp, suffix):
        self.package = package
        self.lp = lp
        self.suffix = suffix
        self._series_cache = {}
        self._sp_cache = {}
        self._config_stack = iconfig.PackageStack(package)

    def _get_host_name(self):
        """Return the configured hostname"""
        return self._config_stack.get('udd.branch_hostname')

    def _get_branch_owner(self):
        """Return the configured owner as a Launchpad or unix username."""
        return self._config_stack.get('udd.branch_owner')

    def _branch_location(self, distro, release, pocket, readonly=False):
        """Returns a tuple (branch_url, is_marked_official)"""
        lp_branch = None
        lp_series = self._get_lp_series(distro, release)
        scheme = "bzr+ssh://"
        if readonly:
            scheme = "http://"
        if lp_series is not None:
            lp_sp = self._get_lp_source_package(distro, release)
            lp_branch = lpapi.lp_call(lp_sp.getBranch, pocket=pocket.title())
        host = self._get_host_name()
        if lp_branch is None:
            distro, suite = self._translate_suite(distro,
                                                  make_suite(release, pocket))
            return ("%s%s/~%s/%s/"
                    "%s/%s/%s%s" % (
                        scheme, host, self._get_branch_owner(),
                        distro, release, self.package, suite, self.suffix),
                    False)
        else:
            path = urlparse.urlparse(lp_branch.self_link).path
            lp_branch_path = path[path[1:].index('/')+1:]
            return ("%s%s%s" % (scheme, host, lp_branch_path),
                    True)

    def side_branch_location(self, distro, release, suite):
        distro, suite = self._translate_suite(distro, suite)
        now = datetime.utcnow()
        now_str = now.strftime("%Y%m%d%H%M")
        return ("bzr+ssh://%s/~%s/%s/%s/%s/%s-%s" % (
            self._get_host_name(),
            self._get_branch_owner(),
            distro, release, self.package, suite, now_str))

    def get_branch(self, importp, possible_transports=None, readonly=False):
        return self.get_branch_parts(
            importp.distro, importp.release, importp.pocket,
            possible_transports=possible_transports, readonly=readonly)

    def get_branch_parts(self, distro, release, pocket,
                         possible_transports=None, readonly=False):
        try:
            return branch.Branch.open(
                self._branch_location(distro, release, pocket,
                                      readonly=readonly)[0],
                possible_transports=possible_transports)
        except errors.NotBranchError:
            return None

    def push_location(self, distro, release, pocket):
        """Returns a tuple (branch_url, is_marked_official)"""
        return self._branch_location(distro, release, pocket)

    def _set_status(self, distro, release, branch, _retry_count=0):
        # series statuses are:
        # Experimental
        # Active Development
        # Pre-release Freeze
        # Current Stable Release
        # Supported
        # Obsolete
        # Future
        #
        # branch statuses are:
        # Experimental
        # Development
        # Mature
        # Merged
        # Abandoned
        series = self._get_lp_series(distro, release)
        if series.status in ('Experimental', 'Active Development',
            'Pre-release Freeze', 'Future'):
            branch.lifecycle_status = 'Development'
        elif series.status in ('Current Stable Release', 'Supported'):
            branch.lifecycle_status = 'Mature'
        elif series.status in ('Obsolete',):
            branch.lifecycle_status = 'Abandoned'
        try:
            lpapi.lp_call(branch.lp_save)
        except HTTPError, e:
            if e.response.status == 412 and _retry_count < 3:
                time.sleep(1)
                self._set_status(distro, release,
                                 lpapi.lp_call(self.lp.load, branch.self_link),
                                 _retry_count=_retry_count+1)
            else:
                raise

    def _set_reviewer(self, branch, _retry_count=0):
        ubuntu_dev = lpapi.lp_call(operator.getitem, self.lp.people,
                                   'ubuntu-dev')
        branch.reviewer = ubuntu_dev
        try:
            lpapi.lp_call(branch.lp_save)
        except HTTPError, e:
            if e.response.status == 412 and _retry_count < 3:
                time.sleep(1)
                self._set_reviewer(lpapi.lp_call(self.lp.load,
                                                 branch.self_link),
                                   _retry_count=_retry_count+1)
            else:
                raise

    def _get_lp_source_package(self, distro, release):
        if (distro, release) in self._sp_cache:
            return self._sp_cache[(distro, release)]
        lp_series = self._get_lp_series(distro, release)
        lp_sp = lpapi.lp_call(lp_series.getSourcePackage, name=self.package)
        self._sp_cache[(distro, release)] = lp_sp
        return lp_sp

    def _get_lp_series(self, distro, release):
        if (distro, release) in self._series_cache:
            return self._series_cache[(distro, release)]
        if distro == "debian" and release in ["woody", "sarge", "etch"]:
            lp_series = None
        else:
            lp_distro = self.lp._udd_distributions[distro]
            lp_series = lpapi.lp_call(lp_distro.getSeries,
                                      name_or_version=release)
        self._series_cache[(distro, release)] = lp_series
        return lp_series

    def _translate_suite(self, distro, suite):
        if distro == "debian":
            if suite == "oldstable":
                suite = "etch"
            elif suite == "stable":
                suite = "lenny"
            elif suite == "testing":
                suite = "squeeze"
            elif suite == "unstable":
                suite = "sid"
        return distro, suite

    def _branch_url(self, distro, release, pocket):
        lp_sp = self._get_lp_source_package(distro, release)
        lp_branch = lpapi.lp_call(lp_sp.getBranch, pocket=pocket.title())
        if lp_branch is None:
            distro, suite = self._translate_suite(
                distro, make_suite(release, pocket))
            return (str(self.lp._root_uri) + '~%s/%s/%s/%s/%s'
                    % (self._get_branch_owner(),
                        distro, release, self.package, suite))
        else:
            return lp_branch.self_link

    def set_official(self, distro, release, pocket, set_status=False,
            set_reviewer=False):
        lp_sp = self._get_lp_source_package(distro, release)
        branch_url = self._branch_url(distro, release, pocket)
        lp_branch = lpapi.lp_call(self.lp.load, branch_url)
        lpapi.lp_call(lp_sp.setBranch, branch=lp_branch, pocket=pocket.title())
        if set_status:
            self._set_status(distro, release, lp_branch)
        if set_reviewer:
            self._set_reviewer(lp_branch)

    def _set_official_direct(self, distro, release, pocket, branch_url):
        lp_sp = self._get_lp_source_package(distro, release)
        if branch_url is not None:
            lp_branch = lpapi.lp_call(self.lp.load, branch_url)
        else:
            lp_branch = None
        lpapi.lp_call(lp_sp.setBranch, branch=lp_branch, pocket=pocket.title())


class LocalShadowBranchStore(BranchStore):

    def __init__(self, package, lp, suffix):
        BranchStore.__init__(self, package, lp, suffix)
        self.branches_dir = os.path.join(conf.get('pi.localbranches_dir'),
                                         package)
        if not os.path.exists(self.branches_dir):
            os.makedirs(self.branches_dir)
            format = bzrdir.format_registry.make_bzrdir('2a')
            to_transport = transport.get_transport(self.branches_dir)
            newdir = format.initialize_on_transport(to_transport)
            repo = newdir.create_repository(shared=True)
            repo.set_make_working_trees(False)

    def side_branch_location(self, distro, release, suite):
        distro, suite = self._translate_suite(distro, suite)
        now = datetime.utcnow()
        now_str = now.strftime("%Y%m%d%H%M")
        return os.path.join(self.branches_dir, "%s-%s-%s"
                            % (distro, suite, now_str))

    def get_branch_parts(self, distro, release, pocket,
                         possible_transports=None, readonly=False):
        local_location = self.push_location(distro, release, pocket)[0]
        try:
            return branch.Branch.open(local_location,
                                      possible_transports=possible_transports)
        except errors.NotBranchError:
            try:
                br_from = branch.Branch.open(
                    self._branch_location(distro, release, pocket,
                                          readonly=readonly)[0],
                    possible_transports=possible_transports)
                br_from.lock_read()
                try:
                    dir = br_from.bzrdir.sprout(local_location)
                    new_br = dir.open_branch()
                    tag._merge_tags_if_possible(br_from, new_br)
                finally:
                    br_from.unlock()
                return new_br
            except errors.NotBranchError:
                return None

    def push_location(self, distro, release, pocket):
        """Returns a tuple (branch_url, is_marked_official)"""
        distro, suite = self._translate_suite(
            distro, make_suite(release, pocket))
        return (os.path.join(self.branches_dir, "%s-%s%s"
                             % (distro, suite, self.suffix)), True)

    def set_official(self, distro, release, pocket, set_status=False,
            set_reviewer=False):
        pass # No-op in this implementation


class MemoryBranchStore(BranchStore):

    def __init__(self, package, lp, suffix):
        self.package = package

    def side_branch_location(self, distro, release, suite):
        distro, suite = self._translate_suite(distro, suite)
        now = datetime.utcnow()
        now_str = now.strftime("%Y%m%d%H%M")
        return os.path.join(conf.get('pi.updates_dir'), self.package,
                suite + "-%s" % now_str)


    def get_branch_parts(self, distro, release, pocket,
            possible_transports=None, readonly=False):
        return None

    def push_location(self, distro, release, pocket):
        raise NotImplementedError(self.push_location)

    def set_official(self, distro, release, pocket, set_status=False,
            set_reviewer=False):
        raise NotImplementedError(self.set_official)


def get_debian_ubuntu_branches(lp, temp_dir, package, bstore,
        ensure_ubuntu_local=False, possible_transports=None):
    if not os.path.exists(temp_dir):
        return (None, None, None)
    try:
        debian_b = branch.Branch.open(os.path.join(temp_dir,
                    default_debian_diff_release))
    except errors.NotBranchError:
        debian_b = bstore.get_branch_parts(u"debian",
                default_debian_diff_release, u"release",
                possible_transports=possible_transports)
    if debian_b is None:
        # Not in unstable
        return (None, None, None)
    ubuntu_tree = None
    ubuntu_current_series = lp.distributions['ubuntu'].current_series.name
    try:
        ubuntu_b = branch.Branch.open(os.path.join(temp_dir,
                    ubuntu_current_series))
        ubuntu_tree = ubuntu_b.bzrdir.open_workingtree()
    except errors.NotBranchError:
        ubuntu_b = bstore.get_branch_parts(u"ubuntu", ubuntu_current_series,
                u"release", possible_transports=possible_transports)
        if ubuntu_b is not None and ensure_ubuntu_local:
            to_transport = transport.get_transport(os.path.join(temp_dir,
                        ubuntu_current_series))
            old_ubuntu_b = ubuntu_b
            old_ubuntu_b.lock_read()
            try:
                revid = old_ubuntu_b.last_revision()
                to_transport.mkdir('.')
                abzrdir = old_ubuntu_b.bzrdir.sprout(to_transport.base,
                        revid, possible_transports=possible_transports)
                ubuntu_b = abzrdir.open_branch()
                tag._merge_tags_if_possible(old_ubuntu_b, ubuntu_b)
                ubuntu_tree = abzrdir.open_workingtree()
            finally:
                old_ubuntu_b.unlock()
    if ubuntu_b is None:
        # Not in Ubuntu
        return (None, None, None)
    return (debian_b, ubuntu_b, ubuntu_tree)


def generate_debian_diffs(lp, temp_dir, package, bstore,
                          possible_transports=None):
    debian_diffs_dir = conf.get('pi.debian_diffs_dir')
    ensure_directory(debian_diffs_dir)
    target_path = os.path.join(debian_diffs_dir, package)
    debian_b, ubuntu_b, _ = get_debian_ubuntu_branches(lp, temp_dir, package,
            bstore, possible_transports=possible_transports)
    if debian_b is None:
        if os.path.exists(target_path):
            os.unlink(target_path)
        return
    debian_tree = debian_b.basis_tree()
    ubuntu_tree = ubuntu_b.basis_tree()
    out = StringIO.StringIO()
    diff.show_diff_trees(debian_tree, ubuntu_tree, out, old_label='debian/',
            new_label='ubuntu/')
    if len(out.getvalue().strip()) > 0:
        with open(target_path, 'wb') as f:
            f.write(out.getvalue())
    else:
        if os.path.exists(target_path):
            os.unlink(target_path)


def possibly_generate_debian_diffs(lp, temp_dir, package, bstore,
                                   possible_transports=None):
    debian_diffs_dir = conf.get('pi.debian_diffs_dir')
    if os.path.exists(os.path.join(debian_diffs_dir, package)):
        return
    generate_debian_diffs(lp, temp_dir, package, bstore,
                          possible_transports=possible_transports)


def generate_ubuntu_merges(lp, temp_dir, package, bstore,
                           possible_transports=None):
    debian_b, ubuntu_b, ubuntu_tree = get_debian_ubuntu_branches(
        lp, temp_dir, package, bstore, ensure_ubuntu_local=True,
        possible_transports=possible_transports)
    ubuntu_merges_dir = conf.get('pi.ubuntu_merges_dir')
    ensure_directory(ubuntu_merges_dir)
    target_path = os.path.join(ubuntu_merges_dir, package)
    if debian_b is None:
        if os.path.exists(target_path):
            os.unlink(target_path)
        return
    debian_b.lock_read()
    try:
        ubuntu_tree.lock_write()
        try:
            out = StringIO.StringIO()
            debian_revid = debian_b.last_revision()
            ubuntu_revid = ubuntu_b.last_revision()
            try:
                merger = merge.Merger.from_revision_ids(None, ubuntu_tree,
                        debian_revid, other_branch=debian_b)
            except errors.UnrelatedBranches:
                out.write("Unrelated branches!")
                return
            merger.merge_type = merge.Merge3Merger
            if (merger.base_rev_id == merger.other_rev_id
                    and merger.other_rev_id is not None):
                if os.path.exists(target_path):
                    os.unlink(target_path)
                return
            if (merger.base_rev_id == ubuntu_revid):
                if os.path.exists(target_path):
                    os.unlink(target_path)
                return
            tree_merger = merger.make_merger()
            tt = tree_merger.make_preview_transform()
            try:
                result_tree = tt.get_preview_tree()
                tree_delta = result_tree.changes_from(merger.this_tree)
                delta.report_delta(out, tree_delta)
                if len(tree_merger.cooked_conflicts) > 0:
                    out.write("conflicts:\n")
                    for conflict in tree_merger.cooked_conflicts:
                        out.write('  %s\n' % conflict)
                out.write("\n\n")
                # TODO: better labels
                old_label = "ubuntu-current/"
                new_label = "ubuntu-merged/"
                out2 = StringIO.StringIO()
                diff.show_diff_trees(merger.this_tree, result_tree, out2,
                        old_label=old_label, new_label=new_label)
                with open(target_path, 'wb') as f:
                    f.write(out.getvalue())
                    f.write(out2.getvalue())
            finally:
                tt.finalize()
        finally:
            ubuntu_tree.unlock()
    finally:
        debian_b.unlock()


def possibly_generate_ubuntu_merges(lp, temp_dir, package, bstore,
                                    possible_transports=None):
    ubuntu_merges_dir = conf.get('pi.ubuntu_merges_dir')
    if os.path.exists(os.path.join(ubuntu_merges_dir, package)):
        return
    generate_ubuntu_merges(lp, temp_dir, package, bstore,
                           possible_transports=possible_transports)


def find_earliest_merge(b, revid):
    for (stop_revid, depth, revno, eom) in b.iter_merge_sorted_revisions(
        start_revision_id=b.last_revision(), stop_revision_id=revid,
        stop_rule='include', direction='forward'):
        if depth == 0:
            return stop_revid
    return None


def load_explanations():
    explanations_file = conf.get('pi.explanations_file')
    explanations = {}
    if not os.path.exists(explanations_file):
        return explanations

    reasons = []
    explanation = ""

    def explain_reasons():
        for reason in reasons:
            explanations[reason] = explanation

    with open(explanations_file) as f:
        for line in f:
            # Each line is either a reason (the exception followed by a
            # traceback signature) or part of an explanation
            line = line.rstrip()
            # Explanation lines begins with at least one space
            if line.startswith(" "):
                # An explanation can span multiple lines.
                explanation += line + '\n'
            else:
                # A reason
                if explanation != "":
                    # The previous explanation is complete
                    explain_reasons()
                    # Be ready for the next reason and its associated
                    # explanation
                    explanation = ""
                    reasons = []
                # Several reasons can share the same explanation
                reasons.append(line)
        if explanation != "":
            explain_reasons()
    return explanations


def update_list(release, component, source_url):
    lists_dir = conf.get('pi.lists_dir')
    target_dir = os.path.join(lists_dir, 'dists', release, component, "source")
    ensure_directory(target_dir)
    target_path = os.path.join(target_dir, "Sources.gz")
    target_lcfile = target_path + ".lc"
    last_check = None
    if os.path.exists(target_lcfile):
        with open(target_lcfile) as f:
            line = f.read().strip()
            try:
                last_check = int(line)
            except ValueError:
                last_check = None
    if last_check is not None:
        now = time.time()
        if now < last_check + (60*60):
            return
    target_tsfile = target_path + ".ts"
    ts = None
    if os.path.exists(target_tsfile):
        with open(target_tsfile) as f:
            ts = f.read().strip()
    req = urllib2.Request(source_url)
    if ts is not None:
        req.add_header('If-Modified-Since', ts)
    opener = urllib2.build_opener()
    try:
        resp = opener.open(req)
    except urllib2.HTTPError, e:
        if e.code != 304:
            raise
    except urllib2.URLError, e:
        print "Error contacting %s" % source_url
        raise
    else:
        with open(target_path, "wb") as f:
            shutil.copyfileobj(resp, f)
        ts = resp.headers.get('Last-Modified')
        with open(target_tsfile, "wb") as f:
            f.write("%s\n" % ts)
        with open(target_lcfile, "wb") as f:
            f.write("%d\n" % time.time())


def refresh_possible_transports(possible_transports):
    """Prune stale connections from possible_transports.

    Mutates possible_transports in place.
    """
    if possible_transports is None:
        return possible_transports
    still_usable = []
    for t in possible_transports:
        try:
            t.has('file-that-does-not-exist')
        except errors.PathError:
            # NosuchFile, PermissionDenied are expected they all ensure that
            # the transport is valid and able to report an exception related to
            # 'file-that-does-not-exist'
            pass
        except errors.TooManyConcurrentRequests:
            mutter(
                "refresh_possible_transports encountered a smart "
                "transport with outstanding request: %r", t)
        except Exception, e:
            # Any other exception is logged for further investigation. But
            # whatever it is, the transport can't be reused so we dropt it.
            mutter("refresh_possible_transports dropping"
                   " transport %r (%s)", t, e)
        else:
            still_usable.append(t)
    possible_transports[:] = still_usable
