import os

from bzrlib import (
    config,
    transport,
    )


_root_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
"""Where the udd branch is rooted at."""


class PkgimportStore(config.TransportIniFileStore):
    """Configuration store for the udd importer.

    The default (no-name) section defines the options common to all imports.
    The named sections apply to the package of the same name.
    """

    def __init__(self, t):
        name = 'pkgimport.conf'
        # If the 'config' dir contains a pkgimport.conf file it takes priority
        new_t = t.clone('config')
        if new_t.has(name):
            t = new_t
        super(PkgimportStore, self).__init__(t, name)

    def get_common_sections(self):
        return config.NameMatcher(self, None).get_sections()


class ImporterStack(config.Stack):
    """Configuration stack for the importer.

    The ``pkgimport.conf`` file defines the configuration options for the
    importer itself, especially all the directories that are used at
    runtime. These options can be overridden in ``locations.conf`` by using a
    section named after the directory used for deployment.
    """

    def __init__(self, possible_transports=None):
        t = transport.get_transport(_root_dir,
                                    possible_transports=possible_transports)
        location_store = config.LocationStore()
        loc_matcher = config.LocationMatcher(location_store, _root_dir)
        pkgimport_store = PkgimportStore(t)
        super(ImporterStack, self).__init__(
            [
                # locations.conf can override any option, based on the tree
                # location
                loc_matcher.get_sections,
                # Otherwise we use pkgimport.conf
                pkgimport_store.get_common_sections,
                ])

    def get(self, name, expand=True, **kwargs):
        """Override base class to expand by default."""
        return super(ImporterStack, self).get(name, expand, **kwargs)


class PackageStack(config.Stack):
    """Configuration stack for a package to be imported.

    Options specific to a package are defined in a section named after the
    package. Failing that, the no-name section defines the options common to
    all packages.
    """

    def __init__(self, package_name, possible_transports=None):
        t = transport.get_transport(_root_dir,
                                    possible_transports=possible_transports)
        pkgimport_store = PkgimportStore(t)
        def package_sections():
            for s in config.NameMatcher(pkgimport_store,
                                        package_name).get_sections():
                yield s
            for s in pkgimport_store.get_common_sections():
                yield s

        super(PackageStack, self).__init__(
            [
                # First we use the package specific section
                package_sections,
                # Otherwise we fallback to the common options
                pkgimport_store.get_common_sections,
                ])

    def get(self, name, expand=True, **kwargs):
        """Override base class to expand by default."""
        return super(PackageStack, self).get(name, expand, **kwargs)

def register_option(*args, **kwargs):
    config.option_registry.register(config.Option(*args, **kwargs))


# 'pi' stands for Package Import and is used as a prefix for all related
# options

register_option('pi.install_dir', default=_root_dir,
                help='Base directory where udd scripts are run from.')

# Common to all scripts:
register_option('pi.base_dir',
                default='/srv/package-import.canonical.com/new',
                help='Base directory for the udd scripts.')
register_option('pi.log_dir', default='{pi.base_dir}/logs',
                help='Base directory for the log files.')
register_option('pi.locks_dir', default='{pi.base_dir}/locks',
                help='Base directory for the lock files.')

config.option_registry.register(
    config.ListOption(
        'pi.distributions', default='debian,ubuntu',
        help='''Which distributions to operate on.

Affect the behaviour of add-import-jobs and list-packages.
'''))

register_option('pi.retry_all_failed_jobs', default='False',
                from_unicode=config.bool_from_store,
                help='''Whether or not to retry all failed jobs.

All failed imports are retried when set. Otherwise only the import jobs that
were currently running the supervisor died are retried.
''')


# Databases (common across multiple components):
register_option('pi.db_type', default='sqlite',
                help='Database type to use, either "sqlite" or "postgres"')
register_option('pi.sqlite_file', default = '{pi.base_dir}/meta.db',
                help='Main database for the package importer.')
register_option('pi.sqlite_package_file', default = '{pi.base_dir}/packages.db',
                help='Packages database for the package importer.')
register_option('pi.sqlite_history_file', default = '{pi.base_dir}/history.db',
                help='History database for the package importer.')
register_option('pi.postgres_host',
                help='Host to connect to postgres.')
register_option('pi.postgres_port',
                help='Port to connect to postgres.')
register_option('pi.postgres_user',
                help='User to connect to postgres.')
register_option('pi.postgres_password',
                help='Password to connect to postgres.')
register_option('pi.postgres_database',
                help='Postgres database to use.')


# Paths common across multiple components:
register_option('pi.package_locks_dir', default = '{pi.locks_dir}/packages',
                help='Directory for package exclusive lock files.')
register_option('pi.script_locks_dir', default = '{pi.locks_dir}/scripts',
                help='Directory for script exclusive lock files.')
register_option('pi.www_dir', default = '{pi.base_dir}/www',
                help='Directory for publishing web pages.')
# The explanations file contains tracebacks - these are highly sensitive to
# changes in the code, so are kept alongside the code.
register_option('pi.explanations_file', default='{pi.install_dir}/explanations',
                help='Explanations for import failures.')


# Used by mass_import:
register_option('pi.import_command',
                default='{pi.install_dir}/bin/import-package {package}',
                help='''Command used to import a package.

{package} is provided when the command is expanded and contains the package
name.
''')
register_option('pi.driver.log_dir', default='{pi.log_dir}/driver',
                help='Directory for the import driver log files.')
register_option('pi.driver.log.debug', default='{pi.driver.log_dir}/debug_log',
                help='Import driver debug log file.')
register_option('pi.driver.log.progress',
                default='{pi.driver.log_dir}/progress_log',
                help='Import package driver progress log file.')
register_option('pi.stop_file', default='{pi.base_dir}/STOP_PLEASE',
                help="A touch'ed file to stop the importer gracefully.")
register_option('pi.max_threads_file', default='{pi.base_dir}/max_threads',
                help='A file containing the maximum number'
                ' of parallel imports.')

# Used by add-import-jobs and list-packages:
register_option('pi.scan_mode', default='source',
                help='''Whether to scan source or binary packages.

Can be either 'source' or 'binary'.  When in 'source' mode, all source
packages will be scanned and 'import_script' will be called for each new or
changed source package.  When in 'binary' mode, all binary packages will be
scanned and 'import_script' will be called for each new or changed binary
package.
''')

# Used by import_package.py:
register_option('pi.packages.log_dir', default='{pi.log_dir}/packages',
                help='Base directory for the log files.')
register_option('pi.lists_dir', default='{pi.base_dir}/lists',
                help='Directory for distribution description mirrors.')
register_option('pi.download_cache_dir', default='{pi.base_dir}/download-cache',
                help='Directory for caching downloaded source packages.')
register_option('pi.localbranches_dir', default='{pi.base_dir}/localbranches',
                help='Directory for caching local branches when debugging.')
register_option('pi.updates_dir', default='{pi.base_dir}/updates',
                help='Working directory to prepare packaging branches.')
register_option('pi.debian_diffs_dir', default='{pi.www_dir}/diffs',
                help='Diffs versus debian versions.')
register_option('pi.ubuntu_merges_dir', default='{pi.www_dir}/merges',
                help='Merges against Ubuntu versions.')

config.option_registry.register(
    config.ListOption(
        'pi.blacklist_distro_releases',
        default=[],
        help='Do not import from these distribution releases'))

# Used by categorize_failures.py:
register_option('pi.web_status_dir', default = '{pi.www_dir}/status',
                help='Directory for publishing importer status.')

# Used by lpapi.py:
# FIXME: Keeping a local credentials file doesn't feel right, when we have a
# newer launchpadlib, we should try to offload this to relying on its APIs.
# --maxb 20110615
register_option('pi.lp_creds_file', default='{pi.base_dir}/lp_creds.txt',
                help='Credentials file to access launchpad.')


# Package specific options
register_option('pi.max_duration',
                from_unicode=config.int_from_store,
                default='86400',
                help='Number of seconds before an import is killed.')
register_option('pi.kill.grace_period',
                from_unicode=config.int_from_store,
                default='10',
                help='Number of seconds before killing an import'
                ' switch from SIGINT to SIGKILL.')
register_option('udd.branch_hostname', default='bazaar.launchpad.net',
                help='Branch hostname.')
register_option('udd.branch_owner', default='ubuntu-branches',
                help='Branch owner as a Launchpad or unix username.')
