import cgi
from datetime import datetime
import os
import random
import simplejson
import sys
import time
import urllib
import urlparse

import httplib2
if os.environ.get('UDD_DEBUG_HTTP'):
    httplib2.debuglevel = 1

from bzrlib.trace import mutter
from bzrlib.plugins.launchpad.lp_api import LAUNCHPAD_API_URLS

from launchpadlib.credentials import Credentials
from launchpadlib.errors import HTTPError
from launchpadlib.launchpad import Launchpad


# Launchpadlib has unfortunately varied the semantics of the service root
# parameter in a slightly complex way over time (see bug 819910).  bzrlib
# has a workaround for it so we'll just use that.
SERVICE_ROOT = LAUNCHPAD_API_URLS['production']


from udd import iconfig
conf = iconfig.ImporterStack()


def get_lp(cache=None):
    lp_creds_file = conf.get('pi.lp_creds_file')
    if os.path.exists(lp_creds_file):
        credentials = Credentials()
        credentials.load(file(lp_creds_file, "r"))
        # Additional parameters, without default values, were inserted into the
        # Launchpad object's __init__ method signature in launchpadlib 1.9.0.
        # To be compatible with old and new versions of launchpadlib, we must
        # detect the correct invocation pattern.
        import inspect
        args = inspect.getargspec(Launchpad.__init__)[0]
        if 'credential_store' in args:
            launchpad = Launchpad(credentials=credentials,
                    authorization_engine=None, credential_store=None,
                    service_root=SERVICE_ROOT, cache=cache)
        else:
            launchpad = Launchpad(credentials=credentials,
                    service_root=SERVICE_ROOT, cache=cache)
        # TODO: handle the case of having credentials that have expired etc
    else:
        launchpad = Launchpad.get_token_and_login('package-import',
                                                  SERVICE_ROOT, cache)
        # TODO: If using a launchpadlib new enough to use gnome-keyring,
        # kwallet, etc. this results in extracting the credentials from the
        # protected store and writing them to disk!
        cred_f = os.fdopen(
            os.open(lp_creds_file, os.O_CREAT|os.O_WRONLY|os.O_TRUNC, 0600),
            'w')
        launchpad.credentials.save(cred_f)
        cred_f.close()
    preload_root_objects(launchpad)
    return launchpad


def lp_call(callable, *args, **kwargs):
    retries = 4
    retried = 0
    while retried < retries:
        try:
            return callable(*args, **kwargs)
        except HTTPError, e:
            retried += 1
            if retried >= retries:
                raise
            if e.response.status // 100 == 5:
                # randomized exponential backoff
                delay = random.randint(10,15)**(retried//2)
                mutter("Sleeping for %d due to %s from LP." % (delay, e))
                time.sleep(delay)
            else:
                if e.response.status == 302:
                    print "Redirected to %s" % e.response.get("location")
                    print "\n".join(["%s: %s"
                                     % pair for pair in e.response.items()])
                    print "\n-----\n%s\n----\n" % e.content
                raise
    assert False, "Shouldn't get here"


def get_collection_slice(collection, start, stop, max_size):
    existing_representation = collection._wadl_resource.representation
    if (existing_representation is not None
        and start < len(existing_representation['entries'])):
        # An optimization: the first page of entries has already
        # been loaded. This can happen if this collection is the
        # return value of a named operation, or if the client did
        # something like check the length of the collection.
        #
        # Either way, we've already made an HTTP request and
        # gotten some entries back. The client has requested a
        # slice that includes some of the entries we already have.
        # In the best case, we can fulfil the slice immediately,
        # without making another HTTP request.
        #
        # Even if we can't fulfil the entire slice, we can get one
        # or more objects from the first page and then have fewer
        # objects to retrieve from the server later. This saves us
        # time and bandwidth, and it might let us save a whole
        # HTTP request.
        entry_page = existing_representation['entries']

        first_page_size = len(entry_page)
        entry_dicts = entry_page[start:stop]
        page_url = existing_representation.get('next_collection_link')
    else:
        # No part of this collection has been loaded yet, or the
        # slice starts beyond the part that has been loaded. We'll
        # use our secret knowledge of lazr.restful to set a value for
        # the ws.start variable. That way we start reading entries
        # from the first one we want.
        first_page_size = max_size + 1
        entry_dicts = []
        page_url = collection._with_url_query_variable_set(
            collection._wadl_resource.url, 'ws.start', start)

    desired_size = stop-start
    more_needed = desired_size - len(entry_dicts)

    # Iterate over pages until we have the correct number of entries.
    while more_needed > 0 and page_url is not None:
        if (first_page_size is not None
                and more_needed > 0 and more_needed < first_page_size):
            # An optimization: it's likely that we need less than
            # a full page of entries, because the number we need
            # is less than the size of the first page we got.
            # Instead of requesting a full-sized page, we'll
            # request only the number of entries we think we'll
            # need. If we're wrong, there's no problem; we'll just
            # keep looping.
            page_url = collection._with_url_query_variable_set(
                page_url, 'ws.size', more_needed)
        representation = simplejson.loads(
            unicode(collection._root._browser.get(page_url)))
        current_page_entries = representation['entries']
        entry_dicts += current_page_entries[:more_needed]
        more_needed = desired_size - len(entry_dicts)

        page_url = representation.get('next_collection_link')
        if page_url is None:
            # We've gotten the entire collection; there are no
            # more entries.
            break
        if first_page_size is None:
            first_page_size = len(current_page_entries)


    # Convert entry_dicts into a list of Entry objects.
    return [resource for resource
            in collection._convert_dicts_to_entries(entry_dicts)]


def iterate_collection(collection, size=20):
    for i in xrange(0, sys.maxint, size):
        slice = get_collection_slice(collection, i, i+size, size)
        if len(slice) == 0:
            break
        for item in slice:
            yield item


def call_with_limited_size(operation, size=20, *args, **kwargs):
    """Invoke the method and process the result."""
    # XXX: Copied from lazr.restfulclient.resource.NamedOperation.__call__ and
    # changed to be a function that takes a NamedOperation, rather than a
    # method on it, also to set the ws.size in the call.
    if len(args) > 0:
        raise TypeError('Method must be called with keyword args.')
    http_method = operation.wadl_method.name
    args = operation._transform_resources_to_links(kwargs)
    request = operation.wadl_method.request

    def _with_url_query_variable_set(url, variable, new_value):
        """A helper method to set a query variable in a URL."""
        parts = urlparse.urlparse(url)
        if parts.query is None:
            params = {}
        else:
            params = cgi.parse_qs(parts.query)
        params[variable] = str(new_value)
        query = urllib.urlencode(params, True)
        return urlparse.urlunparse((parts.scheme, parts.netloc, parts.path,
                    parts.params, query, parts.fragment))

    for key, value in args.items():
        # Certain parameter values should not be JSON-encoded:
        # binary parameters (because they can't be JSON-encoded)
        # and option values (because JSON-encoding them will screw
        # up wadllib's parameter validation). The option value thing
        # is a little hacky, but it's the best solution for now.
        if not isinstance(value, basestring):
            args[key] = simplejson.dumps(value)
    if http_method in ('get', 'head', 'delete'):
        url = operation.wadl_method.build_request_url(**args)
        in_representation = ''
        extra_headers = {}
        url = _with_url_query_variable_set(url, 'ws.size', size)
    else:
        url = operation.wadl_method.build_request_url()
        (media_type,
         in_representation) = operation.wadl_method.build_representation(
            **args)
        extra_headers = { 'Content-type' : media_type }
    response, content = operation.root._browser._request(
        url, in_representation, http_method, extra_headers=extra_headers)

    if response.status == 201:
        return operation._handle_201_response(url, response, content)
    else:
        if http_method == 'post':
            # The method call probably modified this resource in
            # an unknown way. Refresh its representation.
            operation.resource.lp_refresh()
        return operation._handle_200_response(url, response, content)


def preload_root_objects(lp):
    """Cache our own copies of some important Launchpadlib proxy objects.

    Launchpadlib doesn't cache them itself, which causes excessive requests
    (bug 877856) so we keep a few copies here.
    """
    lp._udd_distributions = {}
    for name in ['ubuntu', 'debian']:
        lp._udd_distributions[name] = lp.distributions[name]


class Publication(object):
    """A common interface for source & binary package publishing history."""

    def __init__(self, name, published_time, component_name):
        """Create a `Publication` object.

        :param name: The name of the published package -- source or binary.
        :param published_time: The time when this was published.
        :param component_name: The name of the component this was published
            in.
        """
        self.name = name
        self.published_time = published_time
        self.component_name = component_name

    @classmethod
    def from_source_or_binary(cls, original):
        if getattr(original, 'binary_package_name', None) is None:
            return cls.from_source(original)
        else:
            return cls.from_binary(original)

    @classmethod
    def from_binary(cls, bpph):
        """Create a ``Publication`` from ``BinaryPackagePublishingHistory``.

        :param spph: A ``BinaryPackagePublishingHistory``.
        :return: A ``Publication``.
        """
        name = bpph.binary_package_name
        published_time = datetime.strptime(
            str(bpph.date_created)[:19], "%Y-%m-%d %H:%M:%S")
        return cls(name, published_time, component_name=bpph.component_name)

    @classmethod
    def from_source(cls, spph):
        """Create a ``Publication`` from ``SourcePackagePublishingHistory``.

        :param spph: A ``SourcePackagePublishingHistory``.
        :return: A ``Publication``.
        """
        name = spph.source_package_name
        published_time = datetime.strptime(
            str(spph.date_created)[:19], "%Y-%m-%d %H:%M:%S")
        return cls(name, published_time, component_name=spph.component_name)
