import cgi
import datetime
import hashlib
import os
import re
import shutil
import sys
import time

from udd import (
    icommon,
    iconfig,
    idb,
    )

# Not ideal to declare this here but this is used by several local functions
# below. Refactoring the whole script may help -- vila 2011-12-07
conf = None # Will be set by main()

html_head = '''<html>
<head>
<title>%(title)s</title>
<meta http-equiv="refresh" content="600">
<script src="jquery.min.js"></script>
<script src="jquery.flot.js"></script>
</head>
<body>
<p><a href="http://package-import.ubuntu.com/status/">All Ubuntu Overview</a>,
<a href="http://package-import.ubuntu.com/status/main.html">Ubuntu Main Overview</a></p>

<p>You are looking at information on the Bazaar importer system that serves <a
href="https://wiki.ubuntu.com/DistributedDevelopment">Ubuntu Distributed
Development</a>. It's hoped that you won't have to care about the existence of
this, but things don't always work out that way.</p>

<p>It is expected that uploads appear in the branches in less than 30 minutes,
if it takes longer then something is up and this page should tell you why. To
discuss issues or to ask for a package to be looked at please use our mailing
list <a
href="mailto:ubuntu-distributed-devel@lists.ubuntu.com">ubuntu-distributed-devel@lists.ubuntu.com</a>.</p>

<p>The scripts used to import the packages are available from a <a
href="https://code.launchpad.net/~udd/udd/import-scripts"> bzr branch. </a> If you encounter issues, please
file <a href="https://launchpad.net/udd/+filebug"> bugs</a>. If you're curious, a detailed explanation about how it works is documented in the <a
href="https://wiki.ubuntu.com/DistributedDevelopment/UnderTheHood">
UnderTheHood  </a> part of the <a
href="https://wiki.ubuntu.com/DistributedDevelopment">
wiki.  </a> </p>

<p>This page was last updated at <tt>''' + datetime.datetime.utcnow().isoformat() + ''' UTC.</tt></p>
'''

html_contents = '''<h3>Contents of this page</h3>
<ul>
<li><a href="#stats">Summary</a></li>
<li><a href="#latest">Latest failures</a></li>
<li><a href="#analysis">All failures</a></li>
</ul>
'''

html_tail = '''</body>
</html>
'''


def process_failures():
    db = icommon.StatusDatabase(*idb.get_db_connection(conf, 'main'))
    return db.summarise_failures()


def package_link(package):
    return '<a href="%s.html#%s">%s</a>' % (cgi.escape(package.name, True),
            package.timestamp, cgi.escape(package.name))


def error_page(reason):
    checksum = hashlib.md5()
    checksum.update(reason)
    return "%s.html" % cgi.escape(checksum.hexdigest(), True)


def error_link(reason):
    return '<a href="%s">Packages with this issue</a>' % error_page(reason)


def list_old_pages():
    old_pages = set()
    web_status_dir = conf.get('pi.web_status_dir')
    for page in os.listdir(web_status_dir):
        path = os.path.join(web_status_dir, page)
        if not os.path.isdir(path):
            old_pages.add(path)
    return old_pages


def remove_obsolete_pages(old_pages, new_pages):
    obsolete_pages = old_pages - new_pages
    for path in obsolete_pages:
        if os.path.exists(path) and not os.path.isdir(path):
            os.unlink(path)


def explanation_to_html(explanation_text):
    h = cgi.escape(explanation_text.strip())
    h = re.sub('(https?://[^ ]*)', r'<a href="\1">\1</a>', h)
    return h


def write_individual_pages(reasons, explanations):
    web_status_dir = conf.get('pi.web_status_dir')
    written_paths = set()
    for reason in reasons:
        for info in reasons[reason][:]:
            assert not "/" in info.name
            path = os.path.join(web_status_dir, cgi.escape(info.name,
                        True) + ".html")
            f = open(path, "wb")
            try:
                f.write(html_head % {"title":
                        "Failure reason for %s" % cgi.escape(info.name)})
                f.write("<p>Failed at %s</p>\n" % info.timestamp)
                if reason in explanations:
                    f.write("<p>%s</p>\n" % explanation_to_html(explanations[reason]))
                f.write("<p><pre>\n")
                f.write(cgi.escape(info.raw_reason))
                f.write("\n</pre></p>\n")
                f.write("<div>%s</div>" % error_link(reason))
                f.write(html_tail)
            finally:
                f.close()
            written_paths.add(path)
        path = os.path.join(web_status_dir, error_page(reason))
        f = open(path, "wb")
        try:
            f.write(html_head % {"title": "Failures of type"})
            f.write("<ul>\n")
            for package in reasons[reason][:]:
                f.write("<li>%s</li>\n" % package_link(package))
            f.write("</ul>\n")
            f.write(html_tail)
        finally:
            f.close()
        written_paths.add(path)
    return written_paths


def find_info_where(package_info, clause):
    ret = []
    for info in package_info:
        if clause(info):
            ret.append(info)
    return ret


def plural_form(format_str, quantity, singular, plural):
    """Take a format string which must contain one %d and one %s in that order,
    a numeric quantity, and singular and plural forms relating to it. Compose
    them appropriately."""
    return format_str % (quantity, singular if quantity == 1 else plural)


def write_stats(f, package_info):
    running = find_info_where(package_info, lambda x: x.running)
    running.sort(key=lambda x: x.timestamp)
    outstanding = find_info_where(package_info, lambda x: x.queued)
    outstanding.sort(key=lambda x: x.name)
    failures = find_info_where(package_info,
        lambda x: not x.running and not x.queued)
    fail_normal = find_info_where(failures, lambda x: not x.auto_retry)
    fail_auto_retry = find_info_where(failures,
        lambda x: x.auto_retry and not x.auto_retry_masked)
    fail_masked = find_info_where(failures,
        lambda x: x.auto_retry and x.auto_retry_masked)

    f.write("<ul>\n")

    # The failures summary
    f.write(plural_form("<li>%d %s", len(failures), "failure", "failures"))
    f.write(" - of which:\n")
    f.write("<ul>\n")
    f.write(plural_form("<li>%d normal %s</li>\n",
        len(fail_normal), "failure", "failures"))
    f.write(plural_form("<li>%d spurious %s pending auto-retry</li>\n",
        len(fail_auto_retry), "failure", "failures"))
    f.write(plural_form("<li>%d spurious %s repeated too many times</li>\n",
        len(fail_masked), "failure", "failures"))
    f.write("</ul>\n")
    f.write("</li>\n")

    # The running imports
    f.write("<li>%d currently running\n<ul>\n" % len(running))
    for package in running:
        f.write("<li>%s (since %s)</li>" % (cgi.escape(package.name),
                package.timestamp))
    f.write("</ul>\n</li>\n")

    # The pending imports
    f.write(plural_form("<li>%d outstanding %s\n<ul>\n<li>",
        len(outstanding), "job", "jobs"))
    for package in outstanding:
        f.write(cgi.escape(package.name) + "&nbsp;")
    f.write("</li>\n</ul>\n</li>\n")

    f.write("</ul>\n")


def write_analyis_partial(f, keys, reasons, explanations, clause, text):
    for reason in keys:
        packages = reasons[reason]
        subset = find_info_where(packages, clause)
        if len(subset) > 0:
            f.write("<li>%d packages %s with key %s\n<ul>\n"
                    % (len(subset), text, cgi.escape(reason)))
            if reason in explanations:
                f.write('<li><pre>%s</pre></li>\n'
                        % explanation_to_html(explanations[reason]))
            f.write('<li>')
            subset.sort(key=lambda x: x.name)
            for package in subset:
                f.write(package_link(package) + "&nbsp;")
            f.write("</li>\n</ul>\n</li>\n")


def write_analysis(f, reasons, explanations):
    keys = reasons.keys()
    def compare(a, b):
        return len(reasons[b]) - len(reasons[a])
    keys.sort(cmp=compare)
    write_analyis_partial(f, keys, reasons, explanations,
            lambda x: x.auto_retry_masked, "failed too many times to retry")
    write_analyis_partial(f, keys, reasons, explanations,
            lambda x: not x.auto_retry, "failed")
    write_analyis_partial(f, keys, reasons, explanations,
            lambda x: x.auto_retry and not x.auto_retry_masked,
            "spuriously failed")

def write_latest(f, package_info):
    failed = find_info_where(package_info, lambda x: x.failure_count > 0)
    failed.sort(key=lambda x: x.timestamp, reverse=True)
    i = 0
    for package in failed:
        i += 1
        if i > 50:
            return
        f.write("<li>%s\n<ul>\n" % package_link(package))
        f.write("<li>failed at %s</li>\n" % package.timestamp)
        f.write("<li>with key %s</li>\n" % cgi.escape(package.signature))
        if package.auto_retry_masked:
            f.write("<li>and failed too many times to retry</li>\n")
        elif package.auto_retry:
            f.write("<li>and will be retried automatically around %s</li>\n" % package.auto_retry_time)
        f.write("</ul>\n</li>\n")

def write_overview_page(reasons, package_info, explanations):
    web_status_dir = conf.get('pi.web_status_dir')
    written_paths = set()
    path = os.path.join(web_status_dir, "index.html")
    f = open(path, "wb")
    try:
        f.write(html_head % {"title": "bzr import failures"})
        f.write(html_contents)
        f.write('<div id="stats">\n')
        write_stats(f, package_info)
        f.write('</div>\n<div id="latest">\n<h3>Latest 50 Failures:</h3>\n<ul>\n')
        write_latest(f, package_info)
        f.write('</ul>\n</div>\n')
        f.write('</div>\n<h3>All Failures by Category</h3>\n<div id="analysis">\n<ul>\n')
        write_analysis(f, reasons, explanations)
        f.write('</ul>\n</div>\n')
        db = icommon.HistoryDatabase(*idb.get_db_connection(conf, 'history'))
        write_graphs(f, db.get_counts())
        f.write(html_tail)
    finally:
        f.close()
    written_paths.add(path)
    return written_paths


def get_main_packages():
    db = icommon.PackageDatabase(*idb.get_db_connection(conf, 'package'))
    return db.list_packages_in_main()


def filter_packages(filter_set, reasons, package_info):
    filtered_package_info = find_info_where(package_info,
                                            lambda x: x.name in filter_set)
    filtered_reasons = {}
    for sig in reasons:
        for info in reasons[sig]:
            if info in filtered_package_info:
                filtered_reasons.setdefault(sig, [])
                filtered_reasons[sig].append(info)
    return filtered_reasons, filtered_package_info


def write_graph(f, counts, title, ident):
    f.write('<div>\n%s\n' % title)
    f.write('<div id="%s_graph" style="width:600px;height:300px"></div>\n' % ident)
    f.write('<script type="text/javascript">\n')
    f.write("""$(document).ready(function () {
  $.plot($("#%s_graph"), [[
""" % ident)
    first = True
    for ts, val in counts:
        if not first:
            f.write(",")
        first = False
        f.write("[%s, %d]" % (str(time.mktime(ts.utctimetuple())*1000), val))
    f.write("""]], {xaxis: {mode: "time", min: %s}, yaxis: {min: 0, transform: function (v) { if (v == 0) { return 0; } return Math.log(v); }, inverseTransform: function (v) { if (v == 0) { return 0; } return Math.exp(v); }}});
});
""" % (str(time.mktime(datetime.datetime(2010, 10, 04, 0, 0, 0).utctimetuple())*1000), ))
    f.write('</script>\n')
    f.write('</div>\n')


def write_graphs(f, counts):
    write_graph(f, [(a[0], a[1]) for a in counts], "Number of packages waiting to be imported", "queue")
    write_graph(f, [(a[0], a[2]) for a in counts], "Number of packages that failed to import", "failed")


def write_main_page(reasons, package_info, explanations):
    web_status_dir = conf.get('pi.web_status_dir')
    main_packages = get_main_packages()
    reasons, package_info = filter_packages(main_packages, reasons, package_info)
    written_paths = set()
    path = os.path.join(web_status_dir, "main.html")
    f = open(path, "wb")
    try:
        f.write(html_head % {"title": "bzr import failures for main"})
        f.write(html_contents)
        f.write('<h3>Summary</h3>\n<div id="stats">\n')
        write_stats(f, package_info)
        f.write('</div>\n<div id="latest">\n<h3>Latest 50 Failures:</h3>\n<ul>\n')
        write_latest(f, package_info)
        f.write('</ul>\n</p>\n</div>\n')
        f.write('</div>\n<h3>All Failures by Category</h3>\n<div id="analysis">\n<ul>\n')
        write_analysis(f, reasons, explanations)
        f.write('</ul>\n</div>\n')
        db = icommon.HistoryDatabase(*idb.get_db_connection(conf, 'history'))
        write_graphs(f, db.get_main_counts())
        f.write(html_tail)
    finally:
        f.close()
    written_paths.add(path)
    return written_paths


def copy_js_files():
    web_status_dir = conf.get('pi.web_status_dir')
    source_paths = set(["jquery.min.js", "jquery.flot.js"])
    copied_paths = set()
    for fn in source_paths:
        source = os.path.join(conf.get('pi.install_dir'), fn)
        shutil.copy(source, web_status_dir)
        copied_paths.add(os.path.join(web_status_dir, fn))
    return copied_paths


def get_info():
    reasons, package_info = process_failures()
    explanations = icommon.load_explanations()
    old_pages = list_old_pages()
    new_pages = write_individual_pages(reasons, explanations)
    new_pages |= write_overview_page(reasons, package_info, explanations)
    new_pages |= write_main_page(reasons, package_info, explanations)
    new_pages |= copy_js_files()
    remove_obsolete_pages(old_pages, new_pages)


def install_index():
    web_base_dir = conf.get('pi.www_dir')
    web_status_dir = conf.get('pi.web_status_dir')
    index_path = os.path.join(web_base_dir, 'index.html')
    relpath = os.path.relpath(web_status_dir, web_base_dir)
    content = '''<html>
<head>
<title>Ubuntu Bazaar package importer</title>
<meta http-equiv="refresh" content="0;url=status/">
</head>
<body>
Hi!

See <a href="%s">the status page.</a>
</body>
</html>
''' % (relpath,)
    with open(index_path, 'w') as f:
        f.write(content)


def main():
    global conf
    conf = iconfig.ImporterStack()
    lock = icommon.lock_path(conf.get('pi.script_locks_dir'),
                             'categorise_failures')
    if lock is None:
        print "Another instance of categorise_failures is already running."
        sys.exit(0)
    try:
        icommon.ensure_directory(conf.get('pi.web_status_dir'))
        install_index()
        get_info()
    finally:
        lock.close()
