import codecs
import datetime
import errno
import logging
import os
import Queue
import signal
from stat import ST_DEV, ST_INO
import sys

from bzrlib import cmdline
from launchpadlib.errors import HTTPError

from udd import (
    circuit_breaker,
    icommon,
    iconfig,
    idb,
    threads,
    )


class WatchedFileHandler(logging.FileHandler):
    """
    A handler for logging to a file, which watches the file
    to see if it has changed while in use. This can happen because of
    usage of programs such as newsyslog and logrotate which perform
    log file rotation. This handler, intended for use under Unix,
    watches the file to see if it has changed since the last emit.
    (A file has changed if its device or inode have changed.)
    If it has changed, the old file stream is closed, and the file
    opened to get a new stream.

    This handler is not appropriate for use under Windows, because
    under Windows open files cannot be moved or renamed - logging
    opens the files with exclusive locks - and so there is no need
    for such a handler. Furthermore, ST_INO is not supported under
    Windows; stat always returns zero for this value.

    This handler is based on a suggestion and patch by Chad J.
    Schroeder.
    """
    def __init__(self, filename, mode='a', encoding=None):
        logging.FileHandler.__init__(self, filename, mode, encoding)
        if not os.path.exists(self.baseFilename):
            self.dev, self.ino = -1, -1
        else:
            stat = os.stat(self.baseFilename)
            self.dev, self.ino = stat[ST_DEV], stat[ST_INO]
        self.encoding = encoding

    def _open(self):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        """
        if self.encoding is None:
            stream = open(self.baseFilename, self.mode)
        else:
            stream = codecs.open(self.baseFilename, self.mode, self.encoding)
        return stream


    def emit(self, record):
        """
        Emit a record.

        First check if the underlying file has changed, and if it
        has, close the old stream and reopen the file to get the
        current stream.
        """
        if not os.path.exists(self.baseFilename):
            stat = None
            changed = 1
        else:
            stat = os.stat(self.baseFilename)
            changed = (stat[ST_DEV] != self.dev) or (stat[ST_INO] != self.ino)
        if changed and self.stream is not None:
            self.stream.close()
            self.stream = self._open()
            if stat is None:
                stat = os.stat(self.baseFilename)
            self.dev, self.ino = stat[ST_DEV], stat[ST_INO]
        logging.FileHandler.emit(self, record)


def create_logger():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    conf = iconfig.ImporterStack()
    log_debug_path = conf.get('pi.driver.log.debug')
    log_progress_path = conf.get('pi.driver.log.progress')
    for p in (log_debug_path, log_progress_path):
        dir_path = os.path.dirname(p)
        icommon.ensure_directory(dir_path)

    debug_handler = WatchedFileHandler(log_debug_path)
    progress_handler = WatchedFileHandler(log_progress_path)

    debug_handler.setLevel(logging.DEBUG)
    progress_handler.setLevel(logging.INFO)

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    debug_handler.setFormatter(formatter)
    progress_handler.setFormatter(formatter)

    logger.addHandler(debug_handler)
    logger.addHandler(progress_handler)
    return logger


def report_exception(logger, component, exc):
    import traceback
    exc_class, exc_value, exc_tb = exc
    logger.info('%s failed with exception:\n%s'
                % (component,
                   ''.join(traceback.format_exception(exc_class, exc_value,
                                                      exc_tb))))

class Importer(threads.SubprocessMonitor):

    def __init__(self, logger, package_name, job_id, _command=None):
        """Spawn and monitor an import.

        :param package_name: The name of the package to import (for display
            purposes).

        :param job_id: The status db job id for this import.
        """
        imp_conf = iconfig.ImporterStack()
        if _command is None:
            _command = imp_conf.get('pi.import_command', expand=False)
        _command = imp_conf.expand_options(_command,
                                           env=dict(package=package_name))
        conf = iconfig.PackageStack(package_name)
        super(Importer, self).__init__(
            cmdline.split(_command),
            max_duration=conf.get('pi.max_duration'),
            grace_period=conf.get('pi.kill.grace_period'))
        self.logger = logger
        self.package_name = package_name
        self.job_id = job_id
        self.status_db = icommon.StatusDatabase(*idb.get_db_connection(conf, 'main'))
        self.success = None
        self.failure_sig = None

    def spawn(self):
        self.logger.info("Trying %s" % self.package_name)
        super(Importer, self).spawn()

    def collect(self, timeout=None):
        super(Importer, self).collect(timeout)
        if self.retcode == icommon.no_lock_returncode:
            self.logger.info("Couldn't lock %s, skipping" % self.package_name)
        else:
            # We mostly care about self.err which contains the traceback here
            output = self.out + self.err
            unicode_output = output.decode("utf-8", "replace")
            ascii_output = unicode_output.encode("ascii", "replace")
            self.success = self.retcode == 0
            if self.success:
                self.logger.info("Success %s: %s"
                            % (self.package_name,
                               ascii_output.replace("\n", " ")))
            else:
                self.logger.warning(
                    "Importing %s failed:\n%s" % (
                        self.package_name, ascii_output))
            self.status_db.finish_job(
                self.package_name, self.job_id, self.success, ascii_output)
            if not self.success:
                reason = self.status_db.failure_reason(self.package_name)
                self.failure_sig = self.status_db.failure_signature(reason)
        self.logger.info("thread for %s finished" % self.package_name)


class AllQueue(Queue.Queue):
    """A Queue that always returns a package.

    The jobs in the status db are tried first.

    When no more jobs are pending, the package db is used, even if we are not
    sure that it is needed.
    """

    def __init__(self, logger):
        """Construct an `AllQueue`.

        :param script_path: The path to the script to run to import the
            package.
        """
        Queue.Queue.__init__(self)
        self.logger = logger
        self.tried = set()
        conf = iconfig.ImporterStack()
        self.packages_db = icommon.PackageDatabase(*idb.get_db_connection(conf, 'package'))
        self.status_db = icommon.StatusDatabase(*idb.get_db_connection(conf, 'main'))

    def get_nowait(self):
        # FIXME: It's bogus to implement get_nowait() this way (without
        # locking) but we don't really care since we have a single consumer
        # and no producer. This should be changed to a different model anyway
        # so different kinds of jobs could be queued -- vila 20110221
        try:
            job_id, package_name = Queue.Queue.get_nowait(self)
        except Queue.Empty:
            self.unfinished_tasks += 1
            job_id, package_name = self.next_job()
        return Importer(self.logger, package_name, job_id)

    def next_job(self):
        while True:
            # jobs first
            job_id, package = self.status_db.next_job()
            if package is not None:
                return job_id, package
            else:
                # Now the packages
                try:
                    to_try = self.packages_db.get_one(self.tried)
                except StopIteration:
                    raise Queue.Empty
                if to_try is not None:
                    self.tried.add(to_try)
                    job_id = self.status_db.start_package(to_try)
                    if job_id is not None:
                        return (job_id, to_try)
                else:
                    self.logger.info("All packages requeued, start again")
                    self.tried = set()

# Constants for launchpad status
LP_UP = 1
LP_DOWN = 2
LP_COMING_BACK = 3


class LaunchpadCircuitBreaker(circuit_breaker.CircuitBreaker):
    """Launchpad circuit breaker.

    Launchpad is down on a regular basis for deployments.

    During the outage, many spurious failures may occur.

    As soon as we see a failure known to be retried automatically, we open the
    breaker and immediately requeue the job.

    We don't start new imports while the breaker is open (grace period for
    launchpad to recover, speficied with 'delay').

    We then start imports again and wait for a first success to close the
    breaker or a new failure to start again.
    """

    def __init__(self, logger, threshold, delay):
        super(LaunchpadCircuitBreaker, self).__init__(threshold, delay)
        self.logger = logger
        conf = iconfig.ImporterStack()
        self.status_db = icommon.StatusDatabase(*idb.get_db_connection(conf, 'main'))
        # For log purposes, not to be confused with the breaker state itself
        # which is authoritative.
        self.lp_state = LP_UP

    def see_attempt(self):
        submitted = super(LaunchpadCircuitBreaker, self).see_attempt()
        if self.lp_state == LP_UP and not submitted:
            # First attempt after a failure, lp is down
            self.logger.info("Launchpad just went down")
            self.lp_state = LP_DOWN
        elif self.lp_state == LP_DOWN and submitted:
            # First attempt after the delay
            self.logger.info("Testing if Launchpad is back")
            self.lp_state = LP_COMING_BACK
        return submitted

    def see_success(self):
        super(LaunchpadCircuitBreaker, self).see_attempt()
        if self.lp_state == LP_COMING_BACK:
            self.logger.info("Launchpad *is* back")
            self.lp_state = LP_UP

    def see_import(self, imp):
        # Some imports succeeded, other failed.  See if some failures were
        # transient (transient failures are most likely caused by lp
        # downtimes).
        package_name = imp.package_name
        if imp.success:
            self.see_success()
        else:
            # We can't blame launchpad downtimes for all failures, we rely on
            # the ones declared in RETRY_TABLE (created when
            # 'requeue_package.py --auto' is used)
            if self.status_db.known_auto_retry(imp.failure_sig) is not None:
                self.see_failure()
                # We want to retry asap ('priority')
                self.status_db.retry(package_name, priority=True)
                self.logger.info(
                    "Launchpad is down, %s will be re-tried later"
                    % (imp.package_name,))
                self.lp_state = LP_DOWN
            else:
                self.logger.info(
                    "Failure unrelated to Launchpad downtimes for %s"
                    % (imp.package_name,))


class ImportDriver(threads.ThreadDriver):
    """Monitor the ThreadedImporter.

    This includes planning and spawning imports, tracking their success or
    failures and shutting them down.
    """

    def __init__(self, logger, max_threads):
        super(ImportDriver, self).__init__(None, max_threads)
        self.logger = logger
        # Don't initialize any data that require to be initialized in the
        # thread where the driver runs, use before_start for that. __init__
        # runs in the caller thread while before_start runs in the driver
        # thread itself.

    def before_start(self):
        super(ImportDriver, self).before_start()
        self.queue = AllQueue(self.logger)
        # Launchpad downtimes should be less than 5 minutes (300 seconds).
        self.circuit_breaker = LaunchpadCircuitBreaker(self.logger, 0, 300)

    def do_one_step(self):
        # Don't start a new import while the breaker is open and waiting for
        # things to cool down.
        if self.circuit_breaker.see_attempt():
            try:
                super(ImportDriver, self).do_one_step()
            except Exception:
                # The Driver must go on !
                report_exception(self.logger, 'Driver', sys.exc_info())

    def before_stop(self):
        try:
            super(ImportDriver, self).before_stop()
        except Exception:
            # The Driver must go on !
            report_exception(self.logger, 'Driver', sys.exc_info())

    def start_new_thread(self):
        t = super(ImportDriver, self).start_new_thread()
        if t is not None:
            self.logger.info("Starting thread for %s" % (t.package_name,))

    def collect_terminated_threads(self):
        before = len(self.threads)
        collected = super(ImportDriver, self).collect_terminated_threads()
        after = len(self.threads)
        if before != after:
            # Only mention the running threads when we add or remove some
            self.logger.info("threads for %r still active"
                             % [t.package_name for t in self.threads])
        for imp in collected:
            # Delegate to the circuit breaker to intrepret the import
            self.circuit_breaker.see_import(imp)
        return collected

    def check_time_quota(self):
        abusers = super(ImportDriver, self).check_time_quota()
        for (t, killed) in abusers:
            if killed:
                self.logger.warning(
                    "Importing %s killed, timeout %s exceeded"
                    % (t.package_name,
                       datetime.timedelta(seconds=t.max_duration)))


class Stop(Exception):
    pass


class ImportController(object):
    """Monitor the ImportDriver and interpret signals."""

    def __init__(self, logger):
        self.conf = iconfig.ImporterStack()
        self.status_db = icommon.StatusDatabase(*idb.get_db_connection(self.conf, 'main'))
        # Start with an 8-way Driver
        self.driver = ImportDriver(logger, 8)
        self.logger = logger
        self.stop_file = self.conf.get('pi.stop_file')
        self.max_threads_file = self.conf.get('pi.max_threads_file')


    def run(self):
        lock = icommon.lock_path(self.conf.get('pi.base_dir'), 'mass_import')
        if lock is None:
            self.logger.info("Another main process is running")
            raise Stop
        try:
            if self.conf.get('pi.retry_all_failed_jobs'):
                # We have been configured to retry _all_ failed jobs.
                self.status_db.add_failed_jobs()
            else:
                # Re-queue previously interrupted jobs.
                self.status_db.add_jobs_for_interrupted()
            self.driver.start()
            self.driver.started.wait()
            while True:
                # First, check the driver. We don't check more often than the
                # timeout given below. This timeout should be a balance between
                # user expectations and resources usage, there is no point
                # doing all the filesystem calls every milisecond but the user
                # shouldn't have to wait too long to see his modifications
                # taken into account either.
                self.driver.stopped.wait(0.2)
                if self.driver.stopped.isSet():
                    break
                # Then check the graceful stop file
                if (os.path.exists(self.stop_file)
                    and not self.driver.queue_closed.isSet()):
                    self.driver.queue_closed.set()
                    self.logger.info('Driver would not process new requests')
                    continue
                # And the number of threads
                nb_threads = self.get_max_threads()
                if self.driver.max_threads != nb_threads:
                    self.logger.info('Read %d in %s'
                                     % (nb_threads, self.max_threads_file))
                    self.driver.max_threads = nb_threads
            # Catch driver exception if any
            self.report_driver_exception()
            self.driver.join()
            self.logger.info("Finished")
        finally:
            lock.close()

    def get_max_threads(self):
        # By default, keep the existing one
        nb_threads = self.driver.max_threads
        try:
            with open(self.max_threads_file) as f:
                line = f.readline().strip()
                if line:
                   nb_threads = int(line)
        except IOError, e:
            if e.errno != errno.ENOENT:
                self.logger.warning('Error accessing max threads file %s: %s'
                                    % (self.max_threads_file, str(e)))
        return nb_threads

    def report_driver_exception(self):
        if self.driver.exception is not None:
            report_exception(self.logger, 'Driver', self.driver.exception)
            # No need to ever re-raise this exception, we just reported it
            self.driver.exception = None


def die(logger, controller, signum, frame):
    logger.info("Received signal")
    signal.signal(signum, signal.SIG_DFL)

    controller.driver.stop.set()
    logger.info("Waiting for driver to finish")
    controller.driver.stopped.wait()
    controller.report_driver_exception()
    controller.driver.join()
    logger.info("Driver finished: stopping")
    sys.exit(1)


def main():
    logger = create_logger()
    logger.info("Starting up")
    # Some context to ease debug
    logger.info("PATH: %s" % os.environ.get('PATH', None))
    logger.info("CWD: %s" % os.getcwd())

    controller = ImportController(logger)

    # Install signal handlers
    handler = lambda signum, frame: die(logger, controller, signum, frame)
    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)

    # Run the controller
    try:
        controller.run()
    except HTTPError, e:
        logger.critical(e.content)
        raise
    except Exception, e:
        logger.critical(str(e))
        raise
