import time

from bzrlib import tests

from udd import circuit_breaker


class TestCircuitBreaker(tests.TestCase):

    threshold = None
    delay = None

    def setUp(self):
        super(TestCircuitBreaker, self).setUp()
        self.cb = self.get_cb()

    def get_cb(self, threshold=None, delay=None):
        if threshold is None:
            threshold = self.threshold
        if delay is None:
            delay = self.delay
        return circuit_breaker.CircuitBreaker('test', threshold=threshold,
                                              delay=delay)

    def assertClosed(self):
        self.assertTrue(self.cb.closed())

    def assertOpened(self):
        self.assertFalse(self.cb.closed())
        self.assertEquals(circuit_breaker.OPEN, self.cb.state)

    def assertHalfOpened(self):
        self.assertFalse(self.cb.closed())
        self.assertEquals(circuit_breaker.HALF_OPEN,
                          self.cb.state)


class TestStateMachine(TestCircuitBreaker):
    """Test the three tests and three events with a specific circuit breaker.

    We cover all transitions but for a circuit breaker with a 0 threshold and
    and a 0.0 delay to simplify.

    Other threshold/delay combinations are covered elsewhere.
    """

    threshold = 0 # So we don't have to reach it
    delay = 0.0 # So we don't have to wait

    def test_initially_closed(self):
        self.assertClosed()

    # The usual scenario is (state -- event --> state):
    # closed -- failure --> open -- attempt --> half_open -- success --> closed
    # The following tests cover this scenario.

    def test_open_on_failure(self):
        self.assertClosed()
        self.cb.see_failure()
        self.assertOpened()

    def test_half_open_on_attempt(self):
        self.cb.see_failure()
        self.assertOpened()
        self.assertTrue(self.cb.see_attempt())
        self.assertHalfOpened()

    def test_closed_on_success(self):
        self.cb.see_failure()
        self.assertTrue(self.cb.see_attempt())
        self.assertHalfOpened()
        self.cb.see_success()
        self.assertClosed()

    # Special case scenarios

    def test_stay_closed(self):
        """The other events keep the breaker closed."""
        self.assertTrue(self.cb.see_attempt())
        self.assertClosed()
        self.cb.see_success()
        self.assertClosed()

    def test_keep_trying(self):
        """While half opened, further attempts stay in the same state."""
        self.cb.see_failure()
        self.assertTrue(self.cb.see_attempt())
        self.assertHalfOpened()
        self.assertTrue(self.cb.see_attempt())
        self.assertHalfOpened()

    def test_success_while_opened(self):
        self.cb.see_failure()
        self.cb.see_success()
        self.assertHalfOpened()

    def test_no_cigar(self):
        """The attempt fails, back to the open state."""
        self.cb.see_failure()
        self.assertTrue(self.cb.see_attempt())
        self.cb.see_failure()
        self.assertOpened()


class TestThreshold(TestCircuitBreaker):
    """Test a real threshold"""

    threshold = 1 # We need at least one failure to open

    def test_open_on_failure_with_threshold(self):
        self.cb.see_failure()
        self.assertClosed()
        # Reach the threshold
        self.cb.see_failure()
        self.assertOpened()


class TestDelay(TestCircuitBreaker):
    """Test a real delay"""

    delay = 0.1 # Short enough for not slowing down tests too much

    def test_half_open_after_delay(self):
        self.cb.see_failure()
        self.assertFalse(self.cb.see_attempt())
        self.assertOpened()
        time.sleep(self.delay)
        self.assertTrue(self.cb.see_attempt())
        self.assertHalfOpened()
