from datetime import datetime, timedelta

from bzrlib import tests

from udd import icommon
from udd.lpapi import Publication
from udd.tests import Struct


class FindEarliestMergeTests(tests.TestCaseWithTransport):

    def test_tip_revision(self):
        t = self.make_branch_and_tree(".")
        revid = t.commit("one")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, revid))

    def test_a_parent(self):
        t = self.make_branch_and_tree(".")
        revid = t.commit("one")
        t.commit("two")
        t.commit("three")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, revid))

    def test_merged(self):
        t = self.make_branch_and_tree(".")
        t.commit("one")
        t2 = t.bzrdir.sprout("merged").open_workingtree()
        merged_revid = t2.commit("one and a half")
        t.merge_from_branch(t2.branch)
        revid = t.commit("two")
        t.commit("three")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, merged_revid))


class TestCreateImportJobs(tests.TestCase):

    def setUp(self):
        super(TestCreateImportJobs, self).setUp()

    def makePublication(self, name=None, published_time=None):
        if name is None:
            name = self.getUniqueString('package-name')
        if published_time is None:
            published_time = datetime.now()
        return Publication(name, published_time, component_name=None)

    def test_last_published_date_is_latest_from_pubs(self):
        date = datetime(2011, 1, 1)
        pubs = set([self.makePublication(published_time=date + timedelta(1)),
                    self.makePublication(published_time=date + timedelta(2))])
        last_known_published, ignored = icommon.prepare_for_database(
            date, pubs)
        self.assertEqual(date + timedelta(2), last_known_published)

    def test_last_published_date_is_already_latest(self):
        date = datetime(2011, 1, 1)
        pubs = set([self.makePublication(published_time=date - timedelta(1)),
                    self.makePublication(published_time=date - timedelta(2))])
        last_known_published, ignored = icommon.prepare_for_database(
            date, pubs)
        self.assertEqual(date, last_known_published)

    def test_package_names(self):
        date = datetime(2011, 1, 1)
        pubs = set([self.makePublication(name='foo'),
                    self.makePublication(name='bar')])
        ignored, names = icommon.prepare_for_database(date, pubs)
        self.assertEqual(names, set(['foo', 'bar']))


class TestIterPublishedBinaries(tests.TestCase):

    def setUp(self):
        super(TestIterPublishedBinaries, self).setUp()
        self.patch(icommon, '_lp_get_published_binaries',
                   self._fake_get_published_binaries)

    def _fake_get_published_binaries(self, archive, distro_arch_series,
                                     **filters):
        """A fake version of _lp_get_published_binaries.

        Simply returns its parameters as an iterable.  In the real version, it
        returns an iterable of ``BinaryPackagePublishingHistory`` objects.
        """
        return iter([archive, distro_arch_series, filters])

    def test_series_provided(self):
        # If a series is provided then iter_published_binaries will do one
        # call per distro_arch_series and then merge the results.
        status = object()
        since = object()
        archive = object()
        distribution = Struct(main_archive=archive)
        series = Struct(architectures=['foo', 'bar', 'baz'])
        binaries = icommon.iter_published_binaries(
            distribution, status=status, since=since, series=series)
        # We expect multiple calls to _lp_get_published_binaries. The iterable
        # results of these calls will be chained together.
        self.assertEqual(
            [archive, 'foo', {'status': status, 'created_since_date': since},
             archive, 'bar', {'status': status, 'created_since_date': since},
             archive, 'baz', {'status': status, 'created_since_date': since},
             ],
            list(binaries))

    def test_series_not_provided(self):
        # If no series is provided, then iter_published_binaries will do one
        # call per distro_arch_series per series and then merge the results.
        status = object()
        since = object()
        archive = object()
        distribution = Struct(main_archive=archive)
        distribution.series = [
            Struct(architectures=['a1', 'a2', 'a3']),
            Struct(architectures=['b1', 'b2', 'b3']),
            ]
        binaries = icommon.iter_published_binaries(
            distribution, status=status, since=since)
        # We expect multiple calls to _lp_get_published_binaries. The iterable
        # results of these calls will be chained together.
        self.assertEqual(
            [archive, 'a1', {'status': status, 'created_since_date': since},
             archive, 'a2', {'status': status, 'created_since_date': since},
             archive, 'a3', {'status': status, 'created_since_date': since},
             archive, 'b1', {'status': status, 'created_since_date': since},
             archive, 'b2', {'status': status, 'created_since_date': since},
             archive, 'b3', {'status': status, 'created_since_date': since},
             ],
            list(binaries))
