import datetime

from bzrlib import tests


from udd import icommon, idb


class TestStatusDb(tests.TestCase):

    def setUp(self):
        super(TestStatusDb, self).setUp()
        self.db = icommon.StatusDatabase(*idb.get_memory_db_connection())


class TestFailures(TestStatusDb):

    def test_no_failures_in_empty_db(self):
        reasons, package_info = self.db.summarise_failures()
        self.assertEquals({}, reasons)
        self.assertEquals([], package_info)

    def test_no_unemailed_failures_in_empty_db(self):
        self.assertEquals([], self.db.unemailed_failures())

    def test_unemailed_failures(self):
        now = datetime.datetime.utcnow()
        self.db.add_jobs(['foo'], self.db.JOB_TYPE_PRIORITY)
        self.db.finish_job('foo', 1, False, 'some error', when=now)
        failures = self.db.unemailed_failures()
        self.assertEquals([('foo', 'some error', now)], failures)
        self.db.set_failures_emailed(failures)
        self.assertEquals([], self.db.unemailed_failures())

    def test_failure_reason_unknown_package(self):
        self.assertEqual(None, self.db.failure_reason('foo'))

    def test_running_sentinel_when_started(self):
        # When a job is started it appears as failing with a special signature
        self.db.start_package('foo')
        self.assertEquals(icommon.running_sentinel,
                          self.db.failure_reason('foo'))

    # FIXME: start_package is more complex and can return various job ids (to
    # cope with multiple jobs for the same package ?? Needed in early versions
    # while debugging and creating stale entries ??) -- vila 20111020
    def test_start_package(self):
        self.assertEquals(0, self.db.start_package('foo'))
        # If we attempt to start it again, it's a no-op
        self.assertEquals(None, self.db.start_package('foo'))


class TestJobs(TestStatusDb):

    def test_next_job_empty(self):
        self.assertEqual((0, None), self.db.next_job())

    def test_add_jobs_for_interrupted(self):
        self.db.start_package('foo')
        # Ensures the JOB table is still empty
        self.assertEqual((0, None), self.db.next_job())
        self.db.add_jobs_for_interrupted()
        self.assertEqual((1, 'foo'), self.db.next_job())

    def test_add_failed_jobs_empty(self):
        self.db.add_failed_jobs()
        # No failures means no jobs created
        self.assertEqual((0, None), self.db.next_job())

    def test_add_failed_jobs(self):
        self.db.start_package('foo')
        self.db.add_failed_jobs()
        self.assertEqual((1, 'foo'), self.db.next_job())

    def test_finish_job_success(self):
        # FIXME: It looks like finish_job may never delete the corresponding
        # row in JOBS_TABLE if there is no pending failure -- vila 20111020
        self.db.add_jobs(['foo'], self.db.JOB_TYPE_PRIORITY)
        self.db.finish_job('foo', 1, True, 'win')
        self.assertEquals(None, self.db.failure_reason('foo'))
        # FIXME: Check that nothing is left in OLD_FAILURES_TABLE ?
        # -- vila 20111020

    def test_finish_job_failure(self):
        # FIXME: It looks like finish_job never delete the corresponding row in
        # JOBS_TABLE -- vila 20111020
        self.db.add_jobs(['foo'], self.db.JOB_TYPE_PRIORITY)
        self.db.finish_job('foo', 1, False, 'some error')
        self.assertEquals('some error', self.db.failure_reason('foo'))


class TestImport(TestStatusDb):

    def test_last_import_time_empty(self):
        self.assertEqual(None, self.db.last_import_time())

    def test_last_import_time(self):
        now = datetime.datetime.utcnow()
        self.db.add_import_jobs(['foo'], now)
        self.assertEqual(now, self.db.last_import_time())

    def test_last_import_time_sqlite(self):
        # if the last import time was inserted by sqlite3 then it has a
        # different format from the one that storm would insert. Check
        # that the migration is handled ok.
        now = datetime.datetime.utcnow()
        self.db.conn.execute('INSERT into %s values (?)' % self.db.IMPORT_TABLE,
                (now.strftime("%Y-%m-%d %H:%M:%S"),))
        self.assertEqual(now.replace(microsecond=0), self.db.last_import_time())


class TestRetry(TestStatusDb):

    def test_known_auto_retry_empty(self):
        self.assertEquals(None, self.db.known_auto_retry('whatever'))

# FIXME: Need more tests for the following but they are too hard to test for
# now -- vila 20111020

# _known_auto_retry(self, c, sig):
# known_auto_retry(self, sig):
# retry(self, package, force=False, priority=False, auto=False,
# _retry(self, c, package, signature, timestamp, priority=False):
# _attempt_retry(self, c, info):


class TestSignatures(TestStatusDb):

    def assertSignature(self, expected, raw):
        self.assertEquals(expected, self.db.failure_signature(raw))

    def test_running_signature(self):
        # Running imports use a special failure signature
        self.assertSignature(None, icommon.running_sentinel)

    def test_one_line_signature(self):
        self.assertSignature('AssertionError', 'AssertionError: xx')

    def test_multi_line_signature(self):
        # We use a real-life example here, the apparent duplication of method
        # names is due to calls to the base class methods
        self.assertSignature(
            'OperationalError:do_one_step:do_one_step'
            ':collect_terminated_threads'
            ':collect_terminated_threads:collect'
            ':finish_job:_set_failure',
            '''Traceback (most recent call last):
  File "/srv/package-import.canonical.com/new/scripts/mass_import.py", line 234, in do_one_step
    super(ImportDriver, self).do_one_step()
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 2091, in do_one_step
    self.collect_terminated_threads()
  File "/srv/package-import.canonical.com/new/scripts/mass_import.py", line 253, in collect_terminated_threads
    super(ImportDriver, self).collect_terminated_threads()
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 2119, in collect_terminated_threads
    t.collect()
  File "/srv/package-import.canonical.com/new/scripts/mass_import.py", line 162, in collect
    unicode_output.encode("utf-8", "replace"))
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 552, in finish_job
    self._set_failure(c, package, output, now)
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 485, in _set_failure
    % self.FAILURES_TABLE, (package, reason, now))
OperationalError: table failures has 4 columns but 3 values were supplied
''')

    def test_regular_traceback(self):
        self.assertSignature(
                'bzrlib.errors.NoSuchTag:lookup_tag',
                                 '''Traceback (most recent call last):
  File "/usr/lib/python2.5/site-packages/bzrlib/tag.py", line 109, in lookup_tag
    raise errors.NoSuchTag(tag_name)
bzrlib.errors.NoSuchTag: No such tag: upstream-4.6.2
''')

    def test_traceback_with_verbose_execption(self):
        # make sure we get bzrlib.errors.NoFinalPath without being tricked by:
        # the final empty line, nor the 'file-id:' and 'root trans-id' lines.
        self.assertSignature(
            'bzrlib.errors.NoFinalPath:get_path:_determine_path:final_name',
            ''''Traceback (most recent call last):
  File "/usr/lib/python2.5/site-packages/bzrlib/transform.py", line 2368, in get_path
    self._known_paths[trans_id] = self._determine_path(trans_id)
  File "/usr/lib/python2.5/site-packages/bzrlib/transform.py", line 2358, in _determine_path
    name = self.transform.final_name(trans_id)
  File "/usr/lib/python2.5/site-packages/bzrlib/transform.py", line 470, in final_name
    raise NoFinalPath(trans_id, self)
bzrlib.errors.NoFinalPath: No final name for trans_id 'new-18'
file-id: None
root trans-id: 'new-0'

''')



