import errno
import os
import Queue
import signal
import subprocess
import sys
import threading
import time

try:
    from bzrlib import cethread
except ImportError:
    # Use our copy then
    from udd import cethread


class CatchingExceptionThread(cethread.CatchingExceptionThread):
    """Isolating features that needs to be backported to bzrlib.

    None so far.
    """


class SubprocessMonitor(CatchingExceptionThread):

    def __init__(self, args, max_duration=None, grace_period=None):
        """Create a SubprocessMonitor.

        :param args: A list of arguments to give to python.

        :param max_duration: The max number of seconds the command is expected
            to be run (elapsed time).
        """
        self.started = threading.Event()
        self.stopped = threading.Event()
        super(SubprocessMonitor, self).__init__(target=self.spawn,
                                                sync_event=self.started)
        self.args = args
        self.proc_pid = None
        self.retcode = None
        self.out = self.err = None
        self.started_at = self.ended_at = None
        self.max_duration = max_duration
        self.grace_period = grace_period
        self.killed_with = None
        self.killed_at = None

    def spawn(self):
        """Spawn the python command in a subprocess."""
        proc = subprocess.Popen(['python'] + self.args,
                                executable=sys.executable,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                stdin=subprocess.PIPE)
        self.proc_pid = proc.pid
        self.started_at = time.time()
        self.switch_and_set(self.stopped)
        try:
            self.out, self.err = proc.communicate()
            self.ended_at = time.time()
            self.retcode = proc.returncode
        except OSError, e:
            if e.errno in (errno.ECHILD, errno.ESRCH):
                # The process doesn't exist anymore
                pass
            else:
                raise
        self.stopped.set()

    def collect(self, timeout=None):
        """Collect relevant data before discarding the thread.

        This must be called by the controlling thread. Be aware that since this
        calls join() it may raise the pending exception which must be cleared
        if this behaviour is not the intended one.

        Daughter classes can override this method to collect any relevant data
        (including the pending exception) before the thread is discarded.
        """
        self.join(timeout)

    def kill_with_escalation(self, grace_period=None):
        """Attempt to kill the subprocess.

        :param grace_period: A number of seconds.

        When no ``grace_period`` has been specified, signal.SIGKILL is used to
        kill the process.

        When a ``grace_period`` has been specified, when first called,
        ``signal.SIGTERM`` is used to allow the subprocess to handle the kill
        cleanly.  If called repeatedly, this will not try to kill the process
        again until the end of the specified ``grace_period``. When the grace
        period expires, ``signal.SIGKILL`` is used so that subprocesses that
        don't shutdown when receiving ``signal.SIGTERM`` can still be shut
        down.

        This must be called by the controlling thread.
        """
        if grace_period is None:
            grace_period = self.grace_period
        killed = False
        sig = None
        if grace_period is None:
            sig = signal.SIGKILL
        else:
            if self.killed_with is None:
                sig = signal.SIGTERM
            else:
                # We already tried to kill this subprocess, if it didn't die
                # during the grace period, we'll kill it harder
                now = time.time()
                if now - self.killed_at > grace_period:
                    sig = signal.SIGKILL
        if sig is not None:
            killed = self._kill(sig)
        return killed

    def _kill(self, signum):
        killed = False
        # Don't attempt to kill the subprocess if it hasn't been started
        if self.proc_pid is not None:
            try:
                os.kill(self.proc_pid, signum)
                killed = True
                self.killed_with = signum
                self.killed_at = time.time()
            except OSError, e:
                if e.errno in (errno.ECHILD, errno.ESRCH):
                    # The process doesn't exist anymore.
                    pass
                else:
                    raise
        return killed

    def has_exceeded_time_quota(self, duration=None):
        """Is the thread running for too long ?

        :param duration: If specified will override the ``max_duration``
            attribute.
        """
        if duration is None:
            duration = self.max_duration
        if (duration is not None
            and self.started_at is not None and self.ended_at is None):
            # A timeout is defined, the thread has been started but is not
            # finished yet
            if (time.time() - self.started_at) > duration:
                return True
        return False


class ThreadDriver(CatchingExceptionThread):

    def __init__(self, queue, max_threads, target=None):
        """Create a ThreadDriver.

        :param queue: A thread provider.

        :param max_threads: Maximum number of threads allowed to run
            concurrently.

        :param target: What callable should be run in the thread.
        """
        self.started = threading.Event()
        self.stopped = threading.Event()
        self.stop = threading.Event()
        if target is None:
            target = self.drive
        super(ThreadDriver, self).__init__(target=target,
                                           sync_event=self.started)
        self.queue = queue
        self.queue_closed = threading.Event()
        self.max_threads = max_threads
        self.threads = []

    def drive(self):
        self.before_start()
        # From now on the caller will let us run until we stop or die
        self.switch_and_set(self.stopped)
        while True:
            # Calling wait ensures we can be interrupted properly, but we don't
            # want to block so we use a timeout. We don't use 0 though or other
            # threads may not get any cycle otherwise.
            self.stop.wait(0.001)
            if self.stop.isSet():
                break
            # Check for threads exceeding their expected lifetime
            self.check_time_quota()
            self.do_one_step()
        self.before_stop()
        self.stopped.set()

    def before_start(self):
        pass

    def check_time_quota(self):
        """Ensure threads don't exceed their elapsed time quota.

        :returns: A list of threads exceeding their quota and whether they were
            killed or not.
        """
        abusers = []
        for t in self.threads:
            if t.has_exceeded_time_quota():
                # Die you sucker
                killed = t.kill_with_escalation()
                abusers.append((t, killed))
        return abusers

    def do_one_step(self):
        self.collect_terminated_threads()
        if (not self.stop.isSet() and not self.queue_closed.isSet()
            and len(self.threads) < self.max_threads):
            self.start_new_thread()
        if self.queue_closed.isSet() and not self.threads:
            # The queue is closed and there are no more running threads
            self.stop.set()

    def before_stop(self):
        self.kill_remaining_threads()

    def start_new_thread(self):
        # Get a new thread from the queue and start it
        try:
            new_thread = self.queue.get_nowait()
        except Queue.Empty:
            return None
        self.threads.append(new_thread)
        new_thread.start()
        new_thread.started.wait()
        # Tell the queue we're done with this thread
        self.queue.task_done()
        return new_thread

    def collect_terminated_threads(self):
        # If there are terminated threads, collect them
        collected = []
        for t in self.threads[:]:
            if not t.isAlive():
                t.collect()
                self.threads.remove(t)
                collected.append(t)
        return collected

    def kill_remaining_threads(self):
        # Kill all the remaining threads
        for t in self.threads[:]:
            t.kill_with_escalation()
            # Collect the thread whether we could kill it or not
            t.collect()
            self.threads.remove(t)
